# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import Iterator, Union

from databuilder.models.feature.feature_metadata import FeatureMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.timestamp.timestamp_constants import TIMESTAMP_PROPERTY


# Unlike Watermark, which concerns itself with table implementation specifics (like partition),
# Feature_Watermark is more general and does not care how the feature is stored.
class FeatureWatermark(GraphSerializable):
    """
    Represents the high and low timestamp of data in a Feature.
    """
    NODE_LABEL = 'Feature_Watermark'

    TYPE_ATTR = 'watermark_type'

    WATERMARK_FEATURE_RELATION = 'BELONG_TO_FEATURE'
    FEATURE_WATERMARK_RELATION = 'WATERMARK'

    def __init__(self,
                 feature_group: str,
                 feature_name: str,
                 feature_version: str,
                 timestamp: int,
                 wm_type: str = 'high_watermark',
                 ) -> None:
        self.feature_group = feature_group
        self.feature_name = feature_name
        self.feature_version = feature_version
        self.timestamp = timestamp
        self.wm_type = wm_type

        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()

    def __repr__(self) -> str:
        return f'Feature_Watermark({self.wm_type!r}, {self.timestamp!r}, {self.feature_group!r}, ' \
               f'{self.feature_name!r}, {self.feature_version!r})'

    def _get_feature_key(self) -> str:
        return FeatureMetadata.KEY_FORMAT.format(feature_group=self.feature_group,
                                                 name=self.feature_name,
                                                 version=self.feature_version)

    def _get_watermark_key(self) -> str:
        return f'{self._get_feature_key()}/{self.wm_type}'

    def create_next_node(self) -> Union[GraphNode, None]:
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        yield GraphNode(
            key=self._get_watermark_key(),
            label=FeatureWatermark.NODE_LABEL,
            attributes={
                TIMESTAMP_PROPERTY: self.timestamp,
                FeatureWatermark.TYPE_ATTR: self.wm_type,
            }
        )

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        yield GraphRelationship(
            start_key=self._get_feature_key(),
            start_label=FeatureMetadata.NODE_LABEL,
            end_key=self._get_watermark_key(),
            end_label=FeatureWatermark.NODE_LABEL,
            type=FeatureWatermark.FEATURE_WATERMARK_RELATION,
            reverse_type=FeatureWatermark.WATERMARK_FEATURE_RELATION,
            attributes={}
        )



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from datetime import datetime, timedelta
from typing import (
    Any, Callable, Dict, List, Optional, Tuple,
)

from gremlin_python.process import traversal
from gremlin_python.process.traversal import Column, T
from pyhocon import ConfigFactory, ConfigTree  # noqa: F401

from databuilder import Scoped
from databuilder.clients.neptune_client import NeptuneSessionClient
from databuilder.serializers.neptune_serializer import (
    NEPTUNE_CREATION_TYPE_JOB, NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME,
    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME, NEPTUNE_LAST_EXTRACTED_AT_NODE_PROPERTY_NAME,
    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME,
)
from databuilder.task.base_task import Task  # noqa: F401

LOGGER = logging.getLogger(__name__)


class NeptuneStalenessRemovalTask(Task):
    """
    A Specific task that is to remove stale nodes and relations in Neptune.
    It will use "last_extracted_datetime" attribute assigned from FSNeptuneCSVLoader and if "creation_type" is type job
    the one it is getting it from the config, it will regard the node/relation as stale.
    Not all resources are being published by NeptuneCSVPublisher and you can only set specific LABEL of the node or TYPE
    of relation to perform this deletion.
    """

    NEPTUNE_HOST = 'neptune_host'
    AWS_REGION = 'aws_region'
    AWS_ACCESS_KEY = 'aws_access_key'
    AWS_SECRET_ACCESS_KEY = 'aws_secret_access_key'
    AWS_SESSION_TOKEN = 'aws_session_token'
    AWS_ARN = 'aws_arn'

    # Config keys
    TARGET_NODES = "target_nodes"
    TARGET_RELATIONS = "target_relations"
    DRY_RUN = "dry_run"
    # What is the property name that uniquely identifies each node and edge. Typically is T.id for neptune
    GRAPH_LABEL_ID_PROPERTY_NAME = "graph_label_id_property_name"
    # Staleness max percentage. Safety net to prevent majority of data being deleted.
    STALENESS_MAX_PCT = "staleness_max_pct"
    # Staleness max percentage per LABEL/TYPE. Safety net to prevent majority of data being deleted.
    STALENESS_PCT_MAX_DICT = "staleness_max_pct_dict"
    # Sets how old the nodes and relationships can be
    STALENESS_CUT_OFF_IN_SECONDS = "staleness_cut_off_in_seconds"
    DEFAULT_CONFIG = ConfigFactory.from_dict({
        GRAPH_LABEL_ID_PROPERTY_NAME: T.label,
        STALENESS_MAX_PCT: 5,
        TARGET_NODES: [],
        TARGET_RELATIONS: [],
        STALENESS_PCT_MAX_DICT: {},
        DRY_RUN: False
    })

    def get_scope(self) -> str:
        return 'task.remove_stale_data'

    def init(self, conf: ConfigTree) -> None:
        conf = Scoped.get_scoped_conf(conf, self.get_scope()) \
            .with_fallback(conf) \
            .with_fallback(NeptuneStalenessRemovalTask.DEFAULT_CONFIG)
        self.target_nodes = list(set(conf.get_list(NeptuneStalenessRemovalTask.TARGET_NODES)))
        self.target_relations = list(set(conf.get_list(NeptuneStalenessRemovalTask.TARGET_RELATIONS)))
        self.dry_run = conf.get_bool(NeptuneStalenessRemovalTask.DRY_RUN)
        self.staleness_pct = conf.get_int(NeptuneStalenessRemovalTask.STALENESS_MAX_PCT)
        self.staleness_pct_dict = conf.get(NeptuneStalenessRemovalTask.STALENESS_PCT_MAX_DICT)
        self.graph_label_id = conf.get(NeptuneStalenessRemovalTask.GRAPH_LABEL_ID_PROPERTY_NAME)
        self.staleness_cut_off_in_seconds = conf.get_int(NeptuneStalenessRemovalTask.STALENESS_CUT_OFF_IN_SECONDS)
        self.cutoff_datetime = datetime.utcnow() - timedelta(seconds=self.staleness_cut_off_in_seconds)
        self.gremlin_client = NeptuneSessionClient()
        gremlin_client_conf = Scoped.get_scoped_conf(conf, self.gremlin_client.get_scope())
        self.gremlin_client.init(gremlin_client_conf)

    def run(self) -> None:
        """
        First, performs a safety check to make sure this operation would not delete more than a threshold where
        default threshold is 5%. Once it passes a safety check, it will first delete stale nodes, and then stale
        relations.
        :return:
        """
        self.validate()
        if self.dry_run:
            return
        self._delete_stale_relations()
        self._delete_stale_nodes()

    def validate(self) -> None:
        """
        Validation method. Focused on limit the risk on deleting nodes and relations.
         - Check if deleted nodes will be within 5% of total nodes.
        """
        self._validate_node_staleness_pct()
        self._validate_relation_staleness_pct()

    def _delete_stale_nodes(self) -> None:
        filter_properties = [
            (NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME, NEPTUNE_CREATION_TYPE_JOB, traversal.eq),
            (NEPTUNE_LAST_EXTRACTED_AT_NODE_PROPERTY_NAME, self.cutoff_datetime, traversal.lt)
        ]
        self.gremlin_client.delete_nodes(
            filter_properties=filter_properties,
            node_labels=list(self.target_nodes)
        )

    def _delete_stale_relations(self) -> None:
        filter_properties = [
            (NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME, NEPTUNE_CREATION_TYPE_JOB, traversal.eq),
            (NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME, self.cutoff_datetime, traversal.lt),
        ]

        self.gremlin_client.delete_edges(
            filter_properties=filter_properties,
            edge_labels=list(self.target_relations)
        )

    def _validate_staleness_pct(
            self,
            total_records: List[Dict[str, Any]],
            stale_records: List[Dict[str, Any]],
            types: List[str]
    ) -> None:
        total_count_dict = {record['type']: int(record['count']) for record in total_records}

        for record in stale_records:
            type_str = record['type']
            if type_str not in types:
                continue

            stale_count = record['count']
            if stale_count == 0:
                continue

            node_count = total_count_dict[type_str]
            stale_pct = stale_count * 100 / node_count

            threshold = self.staleness_pct_dict.get(type_str, self.staleness_pct)
            if stale_pct >= threshold:
                raise Exception('Staleness percentage of {} is {} %. Stopping due to over threshold {} %'
                                .format(type_str, stale_pct, threshold))

            LOGGER.info(
                'Will be dropping {stale_count} {record_type} records or {stale_pct}% of {record_type} data'.format(
                    stale_count=stale_count,
                    record_type=type_str,
                    stale_pct=stale_pct
                )
            )

    def _validate_node_staleness_pct(self) -> None:
        total_records = self.get_number_of_nodes_grouped_by_label()
        filter_properties = [
            (NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME, NEPTUNE_CREATION_TYPE_JOB, traversal.eq),
            (NEPTUNE_LAST_EXTRACTED_AT_NODE_PROPERTY_NAME, self.cutoff_datetime, traversal.lt)
        ]
        stale_records = self.get_number_of_nodes_grouped_by_label(
            filter_properties=filter_properties
        )
        self._validate_staleness_pct(
            total_records=total_records,
            stale_records=stale_records,
            types=self.target_nodes
        )

    def get_number_of_nodes_grouped_by_label(
            self,
            filter_properties: Optional[List[Tuple[str, Any, Callable]]] = None
    ) -> List[Dict[str, Any]]:
        if filter_properties is None:
            filter_properties = []
        tx = self.gremlin_client.get_graph().V()
        tx = NeptuneSessionClient.filter_traversal(tx, filter_properties)
        tx = tx.groupCount().by(self.graph_label_id).unfold()
        tx = tx.project('type', 'count')
        tx = tx.by(Column.keys)
        tx = tx.by(Column.values)
        return tx.toList()

    def get_number_of_edges_grouped_by_label(
            self,
            filter_properties: Optional[List[Tuple[str, Any, Callable]]] = None
    ) -> List[Dict]:
        if filter_properties is None:
            filter_properties = []
        tx = self.gremlin_client.get_graph().E()
        tx = NeptuneSessionClient.filter_traversal(tx, filter_properties)
        tx = tx.groupCount().by(self.graph_label_id).unfold()
        tx = tx.project('type', 'count')
        tx = tx.by(Column.keys)
        tx = tx.by(Column.values)
        return tx.toList()

    def _validate_relation_staleness_pct(self) -> None:
        total_records = self.get_number_of_edges_grouped_by_label()
        filter_properties = [
            (NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME, NEPTUNE_CREATION_TYPE_JOB, traversal.eq),
            (NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME, self.cutoff_datetime, traversal.lt)
        ]
        stale_records = self.get_number_of_edges_grouped_by_label(
            filter_properties=filter_properties
        )
        self._validate_staleness_pct(
            total_records=total_records,
            stale_records=stale_records,
            types=self.target_relations
        )



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Any, Optional, Union,
)

from amundsen_rds.models import RDSModel
from amundsen_rds.models.user import User as RDSUser

from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.usage.usage import Usage


class DashboardUsage(Usage):
    """
    A model that encapsulate Dashboard usage between Dashboard and User
    """

    def __init__(self,
                 dashboard_group_id: Optional[str],
                 dashboard_id: Optional[str],
                 email: str,
                 view_count: int,
                 should_create_user_node: Optional[bool] = False,
                 product: Optional[str] = '',
                 cluster: Optional[str] = 'gold',
                 **kwargs: Any
                 ) -> None:
        """
        :param dashboard_group_id:
        :param dashboard_id:
        :param email:
        :param view_count:
        :param should_create_user_node: Enable this if it is fine to create/update User node with only with email
        address. Please be advised that other fields will be emptied. Current use case is to create anonymous user.
        For example, Mode dashboard does not provide which user viewed the dashboard and anonymous user can be used
        to show the usage.
        :param product:
        :param cluster:
        :param kwargs:
        """
        self._should_create_user_node = bool(should_create_user_node)
        Usage.__init__(
            self,
            start_label=DashboardMetadata.DASHBOARD_NODE_LABEL,
            start_key=DashboardMetadata.DASHBOARD_KEY_FORMAT.format(
                product=product,
                cluster=cluster,
                dashboard_group=dashboard_group_id,
                dashboard_name=dashboard_id
            ),
            user_email=email,
            read_count=view_count,
        )

    # override superclass for customized _should_create_user_node behavior
    def create_next_node(self) -> Union[GraphNode, None]:
        if self._should_create_user_node:
            return super().create_next_node()
        return None

    # override superclass for customized _should_create_user_node behavior
    def create_next_record(self) -> Union[RDSModel, None]:
        rec = super().create_next_record()
        if isinstance(rec, RDSUser) and not self._should_create_user_node:
            rec = super().create_next_record()
        return rec



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import importlib
from typing import (
    Any, Dict, List, Optional,
)

from gremlin_python.process.graph_traversal import GraphTraversalSource, __
from gremlin_python.process.traversal import (
    Order, T, TextP,
)
from pyhocon import ConfigTree

from databuilder import Scoped
from databuilder.clients.neptune_client import NeptuneSessionClient
from databuilder.extractor.base_extractor import Extractor
from databuilder.models.cluster.cluster_constants import CLUSTER_REVERSE_RELATION_TYPE
from databuilder.models.dashboard.dashboard_chart import DashboardChart
from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.dashboard.dashboard_query import DashboardQuery
from databuilder.models.description_metadata import DescriptionMetadata
from databuilder.models.owner_constants import OWNER_OF_OBJECT_RELATION_TYPE
from databuilder.models.schema.schema_constant import SCHEMA_REVERSE_RELATION_TYPE
from databuilder.models.table_metadata import TableMetadata
from databuilder.models.timestamp.timestamp_constants import LASTUPDATED_RELATION_TYPE, TIMESTAMP_PROPERTY
from databuilder.models.usage.usage_constants import (
    READ_RELATION_COUNT_PROPERTY, READ_RELATION_TYPE, READ_REVERSE_RELATION_TYPE,
)
from databuilder.models.user import User
from databuilder.serializers.neptune_serializer import METADATA_KEY_PROPERTY_NAME


def _table_search_query(graph: GraphTraversalSource, tag_filter: str) -> List[Dict]:
    traversal = graph.V().hasLabel(TableMetadata.TABLE_NODE_LABEL)
    if tag_filter:
        traversal = traversal.has('published_tag', tag_filter)
    traversal = traversal.project(
        'database',
        'cluster',
        'schema',
        'schema_description',
        'name',
        'key',
        'description',
        'last_updated_timestamp',
        'column_names',
        'column_descriptions',
        'total_usage',
        'unique_usage',
        'tags',
        'badges',
        'programmatic_descriptions'
    )
    traversal = traversal.by(
        __.out(
            TableMetadata.TABLE_SCHEMA_RELATION_TYPE
        ).out(SCHEMA_REVERSE_RELATION_TYPE).out(CLUSTER_REVERSE_RELATION_TYPE).values('name')
    )  # database
    traversal = traversal.by(
        __.out(TableMetadata.TABLE_SCHEMA_RELATION_TYPE).out(SCHEMA_REVERSE_RELATION_TYPE).values('name')
    )  # cluster
    traversal = traversal.by(__.out(TableMetadata.TABLE_SCHEMA_RELATION_TYPE).values('name'))  # schema
    traversal = traversal.by(__.coalesce(
        __.out(TableMetadata.TABLE_SCHEMA_RELATION_TYPE).out(
            DescriptionMetadata.DESCRIPTION_RELATION_TYPE
        ).values('description'),
        __.constant('')
    ))  # schema_description
    traversal = traversal.by('name')  # name
    traversal = traversal.by(T.id)  # key
    traversal = traversal.by(__.coalesce(
        __.out(DescriptionMetadata.DESCRIPTION_RELATION_TYPE).values('description'),
        __.constant('')
    ))  # description
    traversal = traversal.by(
        __.coalesce(__.out(LASTUPDATED_RELATION_TYPE).values(TIMESTAMP_PROPERTY), __.constant(''))
    )  # last_updated_timestamp
    traversal = traversal.by(__.out(TableMetadata.TABLE_COL_RELATION_TYPE).values('name').fold())  # column_names
    traversal = traversal.by(
        __.out(TableMetadata.TABLE_COL_RELATION_TYPE).out(
            DescriptionMetadata.DESCRIPTION_RELATION_TYPE
        ).values('description').fold()
    )  # column_descriptions
    traversal = traversal.by(__.coalesce(
        __.outE(READ_REVERSE_RELATION_TYPE).values('read_count'),
        __.constant(0)).sum()
    )  # total_usage
    traversal = traversal.by(__.outE(READ_REVERSE_RELATION_TYPE).count())  # unique_usage
    traversal = traversal.by(
        __.inE(TableMetadata.TAG_TABLE_RELATION_TYPE).outV().values(METADATA_KEY_PROPERTY_NAME).fold()
    )  # tags
    traversal = traversal.by(
        __.out('HAS_BADGE').values('keys').dedup().fold()
    )  # badges
    traversal = traversal.by(
        __.out(DescriptionMetadata.PROGRAMMATIC_DESCRIPTION_NODE_LABEL).values('description').fold()
    )  # programmatic_descriptions
    traversal = traversal.order().by(__.select('name'), Order.asc)
    return traversal.toList()


def _user_search_query(graph: GraphTraversalSource, tag_filter: str) -> List[Dict]:
    traversal = graph.V().hasLabel(User.USER_NODE_LABEL)
    traversal = traversal.has(User.USER_NODE_FULL_NAME)
    if tag_filter:
        traversal = traversal.where('published_tag', tag_filter)
    traversal = traversal.project(
        'email',
        'first_name',
        'last_name',
        'full_name',
        'github_username',
        'team_name',
        'employee_type',
        'manager_email',
        'slack_id',
        'is_active',
        'role_name',
        'total_read',
        'total_own',
        'total_follow'
    )
    traversal = traversal.by('email')  # email
    traversal = traversal.by('first_name')  # first_name
    traversal = traversal.by('last_name')  # last_name
    traversal = traversal.by('full_name')  # full_name
    traversal = traversal.by('github_username')  # github_username
    traversal = traversal.by('team_name')  # team_name
    traversal = traversal.by('employee_type')  # employee_type
    traversal = traversal.by(__.coalesce(
        __.out(User.USER_MANAGER_RELATION_TYPE).values('email'),
        __.constant(''))
    )  # manager_email
    traversal = traversal.by('slack_id')  # slack_id
    traversal = traversal.by('is_active')  # is_active
    traversal = traversal.by('role_name')  # role_name
    traversal = traversal.by(__.coalesce(
        __.outE(READ_RELATION_TYPE).values('read_count'),
        __.constant(0)
    ).sum())  # total_read
    traversal = traversal.by(__.outE(OWNER_OF_OBJECT_RELATION_TYPE).fold().count())  # total_own
    traversal = traversal.by(__.outE('FOLLOWED_BY').fold().count())  # total_follow
    traversal = traversal.order().by(__.select('email'), Order.asc)
    return traversal.toList()


def _dashboard_search_query(graph: GraphTraversalSource, tag_filter: str) -> List[Dict]:
    traversal = graph.V().hasLabel(DashboardMetadata.DASHBOARD_NODE_LABEL)
    traversal = traversal.has('name')
    if tag_filter:
        traversal = traversal.where('published_tag', tag_filter)

    traversal = traversal.project(
        'group_name',
        'name',
        'cluster',
        'description',
        'group_description',
        'group_url',
        'url',
        'uri',
        'last_successful_run_timestamp',
        'query_names',
        'chart_names',
        'total_usage',
        'tags',
        'badges'
    )
    traversal = traversal.by(
        __.out(DashboardMetadata.DASHBOARD_DASHBOARD_GROUP_RELATION_TYPE).values('name')
    )  # group_name
    traversal = traversal.by('name')  # name
    traversal = traversal.by(
        __.out(
            DashboardMetadata.DASHBOARD_DASHBOARD_GROUP_RELATION_TYPE
        ).out(
            DashboardMetadata.DASHBOARD_GROUP_CLUSTER_RELATION_TYPE
        ).values('name')
    )  # cluster
    traversal = traversal.by(__.coalesce(
        __.out(DashboardMetadata.DASHBOARD_DESCRIPTION_RELATION_TYPE).values('description'),
        __.constant('')
    ))  # description
    traversal = traversal.by(__.coalesce(
        __.out(DashboardMetadata.DASHBOARD_DASHBOARD_GROUP_RELATION_TYPE).out(
            DashboardMetadata.DASHBOARD_DESCRIPTION_RELATION_TYPE
        ).values('description'),
        __.constant('')
    ))  # group_description
    traversal = traversal.by(
        __.out(DashboardMetadata.DASHBOARD_DASHBOARD_GROUP_RELATION_TYPE).values('dashboard_group_url')
    )  # group_url
    traversal = traversal.by('dashboard_url')  # dashboard_url
    traversal = traversal.by('key')  # uri

    traversal = traversal.by(
        __.coalesce(
            __.out('EXECUTED').has('key', TextP.endingWith('_last_successful_execution')).values('timestamp'),
            __.constant('')
        )
    )  # last_successful_run_timestamp
    traversal = traversal.by(
        __.out(DashboardQuery.DASHBOARD_QUERY_RELATION_TYPE).values('name').dedup().fold()
    )  # query_names
    traversal = traversal.by(
        __.out(
            DashboardQuery.DASHBOARD_QUERY_RELATION_TYPE
        ).out(DashboardChart.CHART_RELATION_TYPE).values('name').dedup().fold()
    )  # chart_names
    traversal = traversal.by(__.coalesce(
        __.outE(READ_REVERSE_RELATION_TYPE).values(READ_RELATION_COUNT_PROPERTY),
        __.constant(0)
    ).sum())  # total_usage
    traversal = traversal.by(
        __.out('TAGGED_BY').has('tag_type', 'default').values('keys').dedup().fold()
    )  # tags
    traversal = traversal.by(
        __.out('HAS_BADGE').values('keys').dedup().fold()
    )  # badges

    traversal = traversal.order().by(__.select('name'), Order.asc)

    dashboards = traversal.toList()
    for dashboard in dashboards:
        dashboard['product'] = dashboard['uri'].split('_')[0]

    return dashboards


class NeptuneSearchDataExtractor(Extractor):
    """
    Extractor to fetch data required to support search from Neptune's graph database
    """
    QUERY_FUNCTION_CONFIG_KEY = 'query_function'
    QUERY_FUNCTION_KWARGS_CONFIG_KEY = 'query_function_kwargs'
    ENTITY_TYPE_CONFIG_KEY = 'entity_type'
    JOB_PUBLISH_TAG_CONFIG_KEY = 'job_publish_tag'
    MODEL_CLASS_CONFIG_KEY = 'model_class'

    DEFAULT_QUERY_BY_ENTITY = {
        'table': _table_search_query,
        'user': _user_search_query,
        'dashboard': _dashboard_search_query
    }

    def init(self, conf: ConfigTree) -> None:
        self.conf = conf
        self.entity = conf.get_string(NeptuneSearchDataExtractor.ENTITY_TYPE_CONFIG_KEY, default='table').lower()

        if NeptuneSearchDataExtractor.QUERY_FUNCTION_CONFIG_KEY in conf:
            self.query_function = conf.get(NeptuneSearchDataExtractor.QUERY_FUNCTION_CONFIG_KEY)
        else:
            self.query_function = NeptuneSearchDataExtractor.DEFAULT_QUERY_BY_ENTITY[self.entity]

        self.job_publish_tag = conf.get_string(NeptuneSearchDataExtractor.JOB_PUBLISH_TAG_CONFIG_KEY, '')
        self.neptune_client = NeptuneSessionClient()

        neptune_client_conf = Scoped.get_scoped_conf(conf, self.neptune_client.get_scope())
        self.neptune_client.init(neptune_client_conf)

        model_class = conf.get(NeptuneSearchDataExtractor.MODEL_CLASS_CONFIG_KEY, None)
        if model_class:
            module_name, class_name = model_class.rsplit(".", 1)
            mod = importlib.import_module(module_name)
            self.model_class = getattr(mod, class_name)

        self._extract_iter: Optional[Any] = None

    def close(self) -> None:
        self.neptune_client.close()

    def extract(self) -> Optional[Any]:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()

        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def _get_extract_iter(self) -> Any:
        if not hasattr(self, 'results'):
            self.results = self.query_function(self.neptune_client.get_graph(), tag_filter=self.job_publish_tag)

        for result in self.results:
            if hasattr(self, 'model_class'):
                obj = self.model_class(**result)
                yield obj
            else:
                yield result

    def get_scope(self) -> str:
        return 'extractor.neptune_search_data'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Any, Dict, Optional,
)

from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import (
    NODE_KEY, NODE_LABEL, RELATION_END_KEY, RELATION_END_LABEL, RELATION_REVERSE_TYPE, RELATION_START_KEY,
    RELATION_START_LABEL, RELATION_TYPE,
)
from databuilder.publisher.neo4j_csv_publisher import UNQUOTED_SUFFIX


def serialize_node(node: Optional[GraphNode]) -> Dict[str, Any]:
    if node is None:
        return {}

    node_dict = {
        NODE_LABEL: node.label,
        NODE_KEY: node.key
    }
    for key, value in node.attributes.items():
        key_suffix = _get_neo4j_suffix_value(value)
        formatted_key = f'{key}{key_suffix}'
        node_dict[formatted_key] = value
    return node_dict


def serialize_relationship(relationship: Optional[GraphRelationship]) -> Dict[str, Any]:
    if relationship is None:
        return {}

    relationship_dict = {
        RELATION_START_KEY: relationship.start_key,
        RELATION_START_LABEL: relationship.start_label,
        RELATION_END_KEY: relationship.end_key,
        RELATION_END_LABEL: relationship.end_label,
        RELATION_TYPE: relationship.type,
        RELATION_REVERSE_TYPE: relationship.reverse_type,
    }
    for key, value in relationship.attributes.items():
        key_suffix = _get_neo4j_suffix_value(value)
        formatted_key = f'{key}{key_suffix}'
        relationship_dict[formatted_key] = value

    return relationship_dict


def _get_neo4j_suffix_value(value: Any) -> str:
    if isinstance(value, int):
        return UNQUOTED_SUFFIX

    if isinstance(value, bool):
        return UNQUOTED_SUFFIX

    return ''



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import (
    Any, Iterable, Iterator, List, Optional,
)

from pyhocon import ConfigTree

from databuilder import Scoped


class Transformer(Scoped):
    """
    A transformer transforms a record
    """
    @abc.abstractmethod
    def init(self, conf: ConfigTree) -> None:
        pass

    @abc.abstractmethod
    def transform(self, record: Any) -> Any:
        pass


class NoopTransformer(Transformer):
    """
    A no-op transformer
    """

    def init(self, conf: ConfigTree) -> None:
        pass

    def transform(self, record: Any) -> Any:
        return record

    def get_scope(self) -> str:
        pass


class ChainedTransformer(Transformer):
    """
    A chained transformer that iterates transformers and transforms records.
    Transformers implemented using generator functions can yield multiple records,
    which all get passed to the next transformer.
    Returning None from a transformer filters the record out.
    """

    def __init__(self,
                 transformers: Iterable[Transformer],
                 is_init_transformers: Optional[bool] = False) -> None:
        self.transformers = transformers
        self.is_init_transformers = is_init_transformers

    def init(self, conf: ConfigTree) -> None:
        if self.is_init_transformers:
            for transformer in self.transformers:
                transformer.init(Scoped.get_scoped_conf(conf, transformer.get_scope()))

    def transform(self, record: Any) -> Any:
        records = [record]
        for t in self.transformers:
            new_records: List[Any] = []
            for r in records:
                result = t.transform(r)
                # Get all records if the transformer returns an Iterator.
                if isinstance(result, Iterator):
                    new_records += list(result)

                # Filter the record if it is None
                elif result is not None:
                    new_records.append(result)
            records = new_records

        yield from records

    def get_scope(self) -> str:
        return 'transformer.chained'

    def close(self) -> None:
        for t in self.transformers:
            t.close()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0
from typing import (
    Iterator, Optional, Union,
)

from amundsen_rds.models import RDSModel
from amundsen_rds.models.column import ColumnStat as RDSColumnStat

from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata
from databuilder.models.table_serializable import TableSerializable

LABEL = 'Stat'
STAT_RESOURCE_RELATION_TYPE = 'STAT_OF'
RESOURCE_STAT_RELATION_TYPE = 'STAT'


class TableStats(GraphSerializable, TableSerializable):
    """
    Table stats model.
    """

    KEY_FORMAT = '{db}://{cluster}.{schema}' \
                 '/{table}/{stat_name}/'

    def __init__(self,
                 table_name: str,
                 stat_name: str,
                 stat_val: str,
                 is_metric: bool,
                 db: str = 'hive',
                 schema: str = None,
                 cluster: str = 'gold',
                 start_epoch: str = None,
                 end_epoch: str = None
                 ) -> None:
        if schema is None:
            self.schema, self.table = table_name.split('.')
        else:
            self.table = table_name
            self.schema = schema
        self.db = db
        self.start_epoch = start_epoch
        self.end_epoch = end_epoch
        self.cluster = cluster
        self.stat_name = stat_name
        self.stat_val = str(stat_val)
        # metrics are about the table, stats are about the data in a table
        # ex: table usage is a metric
        self.is_metric = is_metric
        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()
        self._record_iter = self._create_record_iterator()

    def create_next_node(self) -> Optional[GraphNode]:
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def get_table_stat_model_key(self) -> str:
        return TableStats.KEY_FORMAT.format(db=self.db,
                                            cluster=self.cluster,
                                            schema=self.schema,
                                            table=self.table,
                                            stat_name=self.stat_name,
                                            is_metric=self.is_metric)

    def get_table_key(self) -> str:
        # no cluster, schema info from the input
        return TableMetadata.TABLE_KEY_FORMAT.format(db=self.db,
                                                     cluster=self.cluster,
                                                     schema=self.schema,
                                                     tbl=self.table)

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        """
        Create a table stat node
        :return:
        """
        node = GraphNode(
            key=self.get_table_stat_model_key(),
            label=LABEL,
            attributes={
                'stat_val': self.stat_val,
                'stat_type': self.stat_name,
                'start_epoch': self.start_epoch,
                'end_epoch': self.end_epoch,
                'is_metric': self.is_metric,
            }
        )
        yield node

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        """
        Create relation map between table stat record with original table
        :return:
        """
        relationship = GraphRelationship(
            start_key=self.get_table_stat_model_key(),
            start_label=LABEL,
            end_key=self.get_table_key(),
            end_label=TableMetadata.TABLE_NODE_LABEL,
            type=STAT_RESOURCE_RELATION_TYPE,
            reverse_type=RESOURCE_STAT_RELATION_TYPE,
            attributes={}
        )
        yield relationship

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        pass


class TableColumnStats(GraphSerializable, TableSerializable):
    """
    Hive column stats model.
    Each instance represents one row of hive watermark result.
    """
    KEY_FORMAT = '{db}://{cluster}.{schema}' \
                 '/{table}/{col}/{stat_type}/'

    def __init__(self,
                 table_name: str,
                 col_name: str,
                 stat_name: str,
                 stat_val: str,
                 start_epoch: str,
                 end_epoch: str,
                 db: str = 'hive',
                 cluster: str = 'gold',
                 schema: str = None
                 ) -> None:
        if schema is None:
            self.schema, self.table = table_name.split('.')
        else:
            self.table = table_name
            self.schema = schema
        self.db = db
        self.col_name = col_name
        self.start_epoch = start_epoch
        self.end_epoch = end_epoch
        self.cluster = cluster
        self.stat_type = stat_name
        self.stat_val = str(stat_val)
        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()
        self._record_iter = self._create_record_iterator()

    def create_next_node(self) -> Optional[GraphNode]:
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def get_column_stat_model_key(self) -> str:
        return TableColumnStats.KEY_FORMAT.format(db=self.db,
                                                  cluster=self.cluster,
                                                  schema=self.schema,
                                                  table=self.table,
                                                  col=self.col_name,
                                                  stat_type=self.stat_type)

    def get_col_key(self) -> str:
        # no cluster, schema info from the input
        return ColumnMetadata.COLUMN_KEY_FORMAT.format(db=self.db,
                                                       cluster=self.cluster,
                                                       schema=self.schema,
                                                       tbl=self.table,
                                                       col=self.col_name)

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        """
        Create a table stat node
        :return:
        """
        node = GraphNode(
            key=self.get_column_stat_model_key(),
            label=LABEL,
            attributes={
                'stat_val': self.stat_val,
                'stat_type': self.stat_type,
                'start_epoch': self.start_epoch,
                'end_epoch': self.end_epoch,
            }
        )
        yield node

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        """
        Create relation map between table stat record with original hive table
        :return:
        """
        relationship = GraphRelationship(
            start_key=self.get_column_stat_model_key(),
            start_label=LABEL,
            end_key=self.get_col_key(),
            end_label=ColumnMetadata.COLUMN_NODE_LABEL,
            type=STAT_RESOURCE_RELATION_TYPE,
            reverse_type=RESOURCE_STAT_RELATION_TYPE,
            attributes={}
        )
        yield relationship

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        record = RDSColumnStat(
            rk=self.get_column_stat_model_key(),
            stat_val=self.stat_val,
            stat_type=self.stat_type,
            start_epoch=self.start_epoch,
            end_epoch=self.end_epoch,
            column_rk=self.get_col_key()
        )
        yield record



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import csv
import logging
import os
import shutil
from csv import DictWriter
from typing import (
    Any, Dict, FrozenSet,
)

from pyhocon import ConfigFactory, ConfigTree

from databuilder.job.base_job import Job
from databuilder.loader.base_loader import Loader
from databuilder.models.table_serializable import TableSerializable
from databuilder.serializers import mysql_serializer
from databuilder.utils.closer import Closer

LOGGER = logging.getLogger(__name__)


class FSMySQLCSVLoader(Loader):
    """
    Write table record CSV file(s) that can be consumed by MySQLCsvPublisher.
    It assumes that the record it consumes is instance of TableSerializable.
    """
    # Config keys
    RECORD_DIR_PATH = 'record_dir_path'
    FORCE_CREATE_DIR = 'force_create_directory'
    SHOULD_DELETE_CREATED_DIR = 'delete_created_directories'

    _DEFAULT_CONFIG = ConfigFactory.from_dict({
        SHOULD_DELETE_CREATED_DIR: True,
        FORCE_CREATE_DIR: False
    })

    def __init__(self) -> None:
        self._record_file_mapping: Dict[Any, DictWriter] = {}
        self._keys: Dict[FrozenSet[str], int] = {}
        self._closer = Closer()

    def init(self, conf: ConfigTree) -> None:
        """
        Initializing FsMySQLCSVLoader by creating directory for record files.
        Note that the directory defined in configuration should not exist.
        :param conf:
        :return:
        """
        conf = conf.with_fallback(FSMySQLCSVLoader._DEFAULT_CONFIG)

        self._record_dir = conf.get_string(FSMySQLCSVLoader.RECORD_DIR_PATH)
        self._delete_created_dir = conf.get_bool(FSMySQLCSVLoader.SHOULD_DELETE_CREATED_DIR)
        self._force_create_dir = conf.get_bool(FSMySQLCSVLoader.FORCE_CREATE_DIR)
        self._create_directory(self._record_dir)

    def _create_directory(self, path: str) -> None:
        """
        Validate directory does not exist, creates it, register deletion of
        created directory function to Job.closer.
        :param path:
        :return:
        """
        if os.path.exists(path):
            if self._force_create_dir:
                LOGGER.info(f'Directory exist. Deleting directory {path}')
                shutil.rmtree(path)
            else:
                raise RuntimeError(f'Directory should not exist: {path}')

        os.makedirs(path)

        def _delete_dir() -> None:
            if not self._delete_created_dir:
                LOGGER.warning(f'Skip Deleting directory {path}')
                return

            LOGGER.info(f'Deleting directory {path}')
            shutil.rmtree(path)

        # Directory should be deleted after publish is finished
        Job.closer.register(_delete_dir)

    def load(self, csv_serializable: TableSerializable) -> None:
        """
        Writes TableSerializable records into CSV files.
        There are multiple CSV files meaning different tables that this method writes.

        Common pattern for table records:
         1. retrieve csv row (a dict where keys represent a header,
         values represent a row)
         2. using this dict to get a appropriate csv writer and write to it.
         3. repeat 1 and 2

        :param csv_serializable:
        :return:
        """
        record = csv_serializable.next_record()
        while record:
            record_dict = mysql_serializer.serialize_record(record)
            table_name = record.__tablename__
            key = (table_name, self._make_key(record_dict))
            file_suffix = '{}_{}'.format(*key)
            record_writer = self._get_writer(record_dict,
                                             self._record_file_mapping,
                                             key,
                                             self._record_dir,
                                             file_suffix)
            record_writer.writerow(record_dict)
            record = csv_serializable.next_record()

    def _get_writer(self,
                    csv_record_dict: Dict[str, Any],
                    file_mapping: Dict[Any, DictWriter],
                    key: Any,
                    dir_path: str,
                    file_suffix: str
                    ) -> DictWriter:
        """
        Finds a writer based on csv record, key.
        If writer does not exist, it's creates a csv writer and update the mapping.

        :param csv_record_dict:
        :param file_mapping:
        :param key:
        :param dir_path:
        :param file_suffix:
        :return:
        """
        writer = file_mapping.get(key)
        if writer:
            return writer

        LOGGER.info(f'Creating file for {key}')

        file_out = open(f'{dir_path}/{file_suffix}.csv', 'w', encoding='utf8')
        writer = csv.DictWriter(file_out, fieldnames=csv_record_dict.keys(),
                                quoting=csv.QUOTE_NONNUMERIC)

        def file_out_close() -> None:
            LOGGER.info(f'Closing file IO {file_out}')
            file_out.close()
        self._closer.register(file_out_close)

        writer.writeheader()
        file_mapping[key] = writer

        return writer

    def close(self) -> None:
        """
        Any closeable callable registered in _closer, it will close.
        :return:
        """
        self._closer.close()

    def get_scope(self) -> str:
        return "loader.mysql_filesystem_csv"

    def _make_key(self, record_dict: Dict[str, Any]) -> int:
        """ Each unique set of record keys is assigned an increasing numeric key """
        return self._keys.setdefault(frozenset(record_dict.keys()), len(self._keys))



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import re
from typing import Dict

from apache_atlas.client.base_client import AtlasClient
from apache_atlas.exceptions import AtlasServiceException
from apache_atlas.model.misc import SearchFilter
from apache_atlas.model.typedef import AtlasTypesDef
from requests import Timeout

from databuilder.types.atlas.types_def import (
    application_schema, bookmark_schema, cluster_schema, column_schema, column_table_relation, dashboard_chart_schema,
    dashboard_execution_schema, dashboard_group_schema, dashboard_query_schema, dashboard_schema, data_owner_schema,
    database_cluster_relation, database_schema, hive_table_partition, lineage_schema, reader_referenceable_relation,
    reader_schema, report_schema, schema_cluster_relation, schema_schema, source_schema, table_partition_schema,
    table_schema, table_schema_relation, table_source_relation, user_reader_relation, user_schema,
)

LOGGER = logging.getLogger(__name__)


# noinspection PyMethodMayBeStatic
class AtlasEntityInitializer:
    def __init__(self, client: AtlasClient) -> None:
        self.driver = client

    def assign_subtypes(self, regex: str, super_type: str) -> None:
        LOGGER.info(f'\nAssigning {super_type} entity to all the subtypes entity definitions with postfix ')
        entities_to_update = []
        entity_defs = self.driver.typedef.get_all_typedefs(search_filter=SearchFilter()).get('entityDefs', [])

        for e in entity_defs:
            if re.compile(regex).match(e.name) is not None:
                LOGGER.info(f'Assigning {e.name} as a subtype of {super_type}')
                e["superTypes"].append(super_type)
                entities_to_update.append(e)

        typedef_dict = {
            "entityDefs": entities_to_update
        }

        self.driver.typedef.update_atlas_typedefs(AtlasTypesDef(attrs=typedef_dict))
        LOGGER.info(f'Assignment of "{super_type}" Entity to existing "{regex}" entities Completed.\n')

    def create_or_update(self, typedef_dict: Dict, info: str, attempt: int = 1) -> None:
        try:
            LOGGER.info(f"Trying to create {info} Entity")
            self.driver.typedef.create_atlas_typedefs(AtlasTypesDef(attrs=typedef_dict))
        except AtlasServiceException:
            LOGGER.info(f"Already Exists, updating {info} Entity")
            try:
                self.driver.typedef.update_atlas_typedefs(AtlasTypesDef(attrs=typedef_dict))
            except Exception:
                # This is a corner case, for Atlas Sample Data
                LOGGER.warn(f"Error updating {info} Entity.", exc_info=True)

        except Timeout:
            # Sometimes on local atlas instance you do get ReadTimeout a lot.
            # This will try to apply definition 3 times and then cancel
            if attempt < 4:
                LOGGER.info("ReadTimeout - Another Try.")
                self.create_or_update(typedef_dict, info, attempt + 1)
            else:
                LOGGER.info(f"ReadTimeout Exception - Cancelling Operation: {attempt}", exc_info=True)
        except Exception:
            LOGGER.info(f"Error creating/updating {info} Entity Definition", exc_info=True)
        finally:
            LOGGER.info(f"Applied {info} Entity Definition")

    def get_schema_dict(self, schema: str) -> Dict:
        return json.loads(schema)

    def create_table_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(table_schema), "Table")

    def create_column_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(column_schema), "Column")

    def create_column_table_relation(self) -> None:
        self.create_or_update(self.get_schema_dict(column_table_relation), "Column <-> Table")

    def create_cluster_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(cluster_schema), "Cluster")

    def create_database_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(database_schema), "Database")

    def create_database_cluster_relation(self) -> None:
        self.create_or_update(self.get_schema_dict(database_cluster_relation), "Database <-> Cluster")

    def create_schema_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(schema_schema), "Schema")

    def create_schema_cluster_relation(self) -> None:
        self.create_or_update(self.get_schema_dict(schema_cluster_relation), "Schema <-> Database")

    def create_table_schema_relation(self) -> None:
        self.create_or_update(self.get_schema_dict(table_schema_relation), "Table <-> Schema")

    def create_user_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(user_schema), "User")

    def create_reader_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(reader_schema), "Reader")

    def create_bookmark_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(bookmark_schema), "Bookmark")

    def create_report_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(report_schema), "Report")

    def create_user_reader_relation(self) -> None:
        self.create_or_update(self.get_schema_dict(user_reader_relation), "User <-> Reader")

    def create_reader_referenceable_relation(self) -> None:
        self.create_or_update(self.get_schema_dict(reader_referenceable_relation), "Reader <-> Referenceable")

    def create_table_partition_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(table_partition_schema), "Partition")

    def create_hive_table_partition(self) -> None:
        self.create_or_update(self.get_schema_dict(hive_table_partition), "Hive Table Partition")

    def create_data_owner_relation(self) -> None:
        self.create_or_update(self.get_schema_dict(data_owner_schema), "Data Owner Relation")

    def create_application_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(application_schema), "Application")

    def create_source_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(source_schema), "Source")

    def create_table_source_relation(self) -> None:
        self.create_or_update(self.get_schema_dict(table_source_relation), "Table <-> Source")

    def create_lineage_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(lineage_schema), "LineageProcess")

    def create_dashboard_group_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(dashboard_group_schema), "Dashboard Group")

    def create_dashboard_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(dashboard_schema), "Dashboard")

    def create_dashboard_chart_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(dashboard_chart_schema), "Dashboard Chart")

    def create_dashboard_query_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(dashboard_query_schema), "Dashboard Query")

    def create_dashboard_execution_schema(self) -> None:
        self.create_or_update(self.get_schema_dict(dashboard_execution_schema), "Dashboard Execution")

    def create_dashboard_cluster_relation(self) -> None:
        self.create_or_update(self.get_schema_dict(database_cluster_relation), "Dashboard <-> Cluster")

    def create_required_entities(self, fix_existing_data: bool = False) -> None:
        """
        IMPORTANT: The order of the entity definition matters.
        Please keep this order.
        :return: Creates or Updates the entity definition in Apache Atlas
        """
        self.create_cluster_schema()
        self.create_column_schema()
        self.create_reader_schema()
        self.create_user_schema()
        self.create_bookmark_schema()
        self.create_report_schema()
        self.create_database_schema()
        self.create_database_cluster_relation()
        self.create_schema_schema()
        self.create_schema_cluster_relation()
        self.create_table_schema()
        self.create_column_table_relation()
        self.create_table_schema_relation()
        self.create_source_schema()
        self.create_table_source_relation()
        self.create_application_schema()
        self.create_lineage_schema()
        self.assign_subtypes(regex="(.*)_table$", super_type="Table")
        self.assign_subtypes(regex="(.*)_column$", super_type="Column")
        self.create_user_reader_relation()
        self.create_reader_referenceable_relation()
        self.create_table_partition_schema()
        self.create_hive_table_partition()
        self.create_data_owner_relation()
        self.create_dashboard_group_schema()
        self.create_dashboard_schema()
        self.create_dashboard_query_schema()
        self.create_dashboard_chart_schema()
        self.create_dashboard_execution_schema()
        self.create_dashboard_cluster_relation()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import (
    Any, Iterator, Optional, Union,
)

from amundsen_common.utils.atlas import AtlasCommonParams, AtlasDashboardTypes
from amundsen_rds.models import RDSModel
from amundsen_rds.models.dashboard import DashboardChart as RDSDashboardChart

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.dashboard.dashboard_query import DashboardQuery
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_serializable import TableSerializable
from databuilder.serializers.atlas_serializer import (
    add_entity_relationship, get_entity_attrs, get_entity_relationships,
)
from databuilder.utils.atlas import AtlasSerializedEntityOperation

LOGGER = logging.getLogger(__name__)


class DashboardChart(GraphSerializable, TableSerializable, AtlasSerializable):
    """
    A model that encapsulate Dashboard's charts
    """
    DASHBOARD_CHART_LABEL = 'Chart'
    DASHBOARD_CHART_KEY_FORMAT = '{product}_dashboard://{cluster}.{dashboard_group_id}/' \
                                 '{dashboard_id}/query/{query_id}/chart/{chart_id}'
    CHART_RELATION_TYPE = 'HAS_CHART'
    CHART_REVERSE_RELATION_TYPE = 'CHART_OF'

    def __init__(self,
                 dashboard_group_id: Optional[str],
                 dashboard_id: Optional[str],
                 query_id: str,
                 chart_id: str,
                 chart_name: Optional[str] = None,
                 chart_type: Optional[str] = None,
                 chart_url: Optional[str] = None,
                 product: Optional[str] = '',
                 cluster: str = 'gold',
                 **kwargs: Any
                 ) -> None:
        self._dashboard_group_id = dashboard_group_id
        self._dashboard_id = dashboard_id
        self._query_id = query_id
        self._chart_id = chart_id if chart_id else chart_name
        self._chart_name = chart_name
        self._chart_type = chart_type
        self._chart_url = chart_url
        self._product = product
        self._cluster = cluster
        self._node_iterator = self._create_node_iterator()
        self._relation_iterator = self._create_relation_iterator()
        self._record_iterator = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()

    def create_next_node(self) -> Union[GraphNode, None]:
        try:
            return next(self._node_iterator)
        except StopIteration:
            return None

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        node_attributes = {
            'id': self._chart_id
        }

        if self._chart_name:
            node_attributes['name'] = self._chart_name

        if self._chart_type:
            node_attributes['type'] = self._chart_type

        if self._chart_url:
            node_attributes['url'] = self._chart_url

        node = GraphNode(
            key=self._get_chart_node_key(),
            label=DashboardChart.DASHBOARD_CHART_LABEL,
            attributes=node_attributes
        )
        yield node

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        try:
            return next(self._relation_iterator)
        except StopIteration:
            return None

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        relationship = GraphRelationship(
            start_label=DashboardQuery.DASHBOARD_QUERY_LABEL,
            start_key=DashboardQuery.DASHBOARD_QUERY_KEY_FORMAT.format(
                product=self._product,
                cluster=self._cluster,
                dashboard_group_id=self._dashboard_group_id,
                dashboard_id=self._dashboard_id,
                query_id=self._query_id
            ),
            end_label=DashboardChart.DASHBOARD_CHART_LABEL,
            end_key=self._get_chart_node_key(),
            type=DashboardChart.CHART_RELATION_TYPE,
            reverse_type=DashboardChart.CHART_REVERSE_RELATION_TYPE,
            attributes={}
        )
        yield relationship

    def _get_chart_node_key(self) -> str:
        return DashboardChart.DASHBOARD_CHART_KEY_FORMAT.format(
            product=self._product,
            cluster=self._cluster,
            dashboard_group_id=self._dashboard_group_id,
            dashboard_id=self._dashboard_id,
            query_id=self._query_id,
            chart_id=self._chart_id
        )

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iterator)
        except StopIteration:
            return None

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        record = RDSDashboardChart(
            rk=self._get_chart_node_key(),
            id=self._chart_id,
            query_rk=DashboardQuery.DASHBOARD_QUERY_KEY_FORMAT.format(
                product=self._product,
                cluster=self._cluster,
                dashboard_group_id=self._dashboard_group_id,
                dashboard_id=self._dashboard_id,
                query_id=self._query_id
            )
        )
        if self._chart_name:
            record.name = self._chart_name
        if self._chart_type:
            record.type = self._chart_type
        if self._chart_url:
            record.url = self._chart_url

        yield record

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        return None

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:
        # Chart
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, self._get_chart_node_key()),
            ('name', self._chart_name),
            ('type', self._chart_type),
            ('url', self._chart_url)
        ]

        chart_entity_attrs = get_entity_attrs(attrs_mapping)

        relationship_list = []  # type: ignore

        add_entity_relationship(
            relationship_list,
            'query',
            AtlasDashboardTypes.query,
            DashboardQuery.DASHBOARD_QUERY_KEY_FORMAT.format(
                product=self._product,
                cluster=self._cluster,
                dashboard_group_id=self._dashboard_group_id,
                dashboard_id=self._dashboard_id,
                query_id=self._query_id
            )
        )

        chart_entity = AtlasEntity(
            typeName=AtlasDashboardTypes.chart,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=chart_entity_attrs,
            relationships=get_entity_relationships(relationship_list)
        )

        yield chart_entity

    def __repr__(self) -> str:
        return f'DashboardChart({self._dashboard_group_id!r}, {self._dashboard_id!r}, ' \
               f'{self._query_id!r}, {self._chart_id!r}, {self._chart_name!r}, {self._chart_type!r}, ' \
               f'{self._chart_url!r}, {self._product!r}, {self._cluster!r})'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from collections import namedtuple
from itertools import groupby
from typing import (
    Any, Dict, Iterator, Union,
)

from pyhocon import ConfigFactory, ConfigTree
from sqlalchemy.engine.url import make_url

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.sql_alchemy_extractor import SQLAlchemyExtractor
from databuilder.extractor.table_metadata_constants import PARTITION_BADGE
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata

TableKey = namedtuple('TableKey', ['schema', 'table_name'])

LOGGER = logging.getLogger(__name__)


class HiveTableMetadataExtractor(Extractor):
    """
    Extracts Hive table and column metadata from underlying meta store database using SQLAlchemyExtractor
    """
    EXTRACT_SQL = 'extract_sql'
    # SELECT statement from hive metastore database to extract table and column metadata
    # Below SELECT statement uses UNION to combining two queries together.
    # 1st query is retrieving partition columns
    # 2nd query is retrieving columns
    # Using UNION to combine above two statements and order by table & partition identifier.
    DEFAULT_SQL_STATEMENT = """
    SELECT source.* FROM
    (SELECT t.TBL_ID, d.NAME as `schema`, t.TBL_NAME name, t.TBL_TYPE, tp.PARAM_VALUE as description,
           p.PKEY_NAME as col_name, p.INTEGER_IDX as col_sort_order,
           p.PKEY_TYPE as col_type, p.PKEY_COMMENT as col_description, 1 as "is_partition_col",
           IF(t.TBL_TYPE = 'VIRTUAL_VIEW', 1, 0) "is_view"
    FROM TBLS t
    JOIN DBS d ON t.DB_ID = d.DB_ID
    JOIN PARTITION_KEYS p ON t.TBL_ID = p.TBL_ID
    LEFT JOIN TABLE_PARAMS tp ON (t.TBL_ID = tp.TBL_ID AND tp.PARAM_KEY='comment')
    {where_clause_suffix}
    UNION
    SELECT t.TBL_ID, d.NAME as `schema`, t.TBL_NAME name, t.TBL_TYPE, tp.PARAM_VALUE as description,
           c.COLUMN_NAME as col_name, c.INTEGER_IDX as col_sort_order,
           c.TYPE_NAME as col_type, c.COMMENT as col_description, 0 as "is_partition_col",
           IF(t.TBL_TYPE = 'VIRTUAL_VIEW', 1, 0) "is_view"
    FROM TBLS t
    JOIN DBS d ON t.DB_ID = d.DB_ID
    JOIN SDS s ON t.SD_ID = s.SD_ID
    JOIN COLUMNS_V2 c ON s.CD_ID = c.CD_ID
    LEFT JOIN TABLE_PARAMS tp ON (t.TBL_ID = tp.TBL_ID AND tp.PARAM_KEY='comment')
    {where_clause_suffix}
    ) source
    ORDER by tbl_id, is_partition_col desc;
    """

    DEFAULT_POSTGRES_SQL_STATEMENT = """
    SELECT source.* FROM
    (SELECT t."TBL_ID" as tbl_id, d."NAME" as "schema", t."TBL_NAME" as name, t."TBL_TYPE",
           tp."PARAM_VALUE" as description, p."PKEY_NAME" as col_name, p."INTEGER_IDX" as col_sort_order,
           p."PKEY_TYPE" as col_type, p."PKEY_COMMENT" as col_description, 1 as "is_partition_col",
           CASE WHEN t."TBL_TYPE" = 'VIRTUAL_VIEW' THEN 1
                ELSE 0 END as "is_view"
    FROM "TBLS" t
    JOIN "DBS" d ON t."DB_ID" = d."DB_ID"
    JOIN "PARTITION_KEYS" p ON t."TBL_ID" = p."TBL_ID"
    LEFT JOIN "TABLE_PARAMS" tp ON (t."TBL_ID" = tp."TBL_ID" AND tp."PARAM_KEY"='comment')
    {where_clause_suffix}
    UNION
    SELECT t."TBL_ID" as tbl_id, d."NAME" as "schema", t."TBL_NAME" as name, t."TBL_TYPE",
           tp."PARAM_VALUE" as description, c."COLUMN_NAME" as col_name, c."INTEGER_IDX" as col_sort_order,
           c."TYPE_NAME" as col_type, c."COMMENT" as col_description, 0 as "is_partition_col",
           CASE WHEN t."TBL_TYPE" = 'VIRTUAL_VIEW' THEN 1
                ELSE 0 END as "is_view"
    FROM "TBLS" t
    JOIN "DBS" d ON t."DB_ID" = d."DB_ID"
    JOIN "SDS" s ON t."SD_ID" = s."SD_ID"
    JOIN "COLUMNS_V2" c ON s."CD_ID" = c."CD_ID"
    LEFT JOIN "TABLE_PARAMS" tp ON (t."TBL_ID" = tp."TBL_ID" AND tp."PARAM_KEY"='comment')
    {where_clause_suffix}
    ) source
    ORDER by tbl_id, is_partition_col desc;
    """

    # CONFIG KEYS
    WHERE_CLAUSE_SUFFIX_KEY = 'where_clause_suffix'
    CLUSTER_KEY = 'cluster'

    DEFAULT_CONFIG = ConfigFactory.from_dict({WHERE_CLAUSE_SUFFIX_KEY: ' ',
                                              CLUSTER_KEY: 'gold'})

    def init(self, conf: ConfigTree) -> None:
        conf = conf.with_fallback(HiveTableMetadataExtractor.DEFAULT_CONFIG)
        self._cluster = conf.get_string(HiveTableMetadataExtractor.CLUSTER_KEY)

        self._alchemy_extractor = SQLAlchemyExtractor()

        sql_alch_conf = Scoped.get_scoped_conf(conf, self._alchemy_extractor.get_scope())
        default_sql = self._choose_default_sql_stm(sql_alch_conf).format(
            where_clause_suffix=conf.get_string(HiveTableMetadataExtractor.WHERE_CLAUSE_SUFFIX_KEY))

        self.sql_stmt = conf.get_string(HiveTableMetadataExtractor.EXTRACT_SQL, default=default_sql)

        LOGGER.info('SQL for hive metastore: %s', self.sql_stmt)

        sql_alch_conf = sql_alch_conf.with_fallback(ConfigFactory.from_dict(
            {SQLAlchemyExtractor.EXTRACT_SQL: self.sql_stmt}))
        self._alchemy_extractor.init(sql_alch_conf)
        self._extract_iter: Union[None, Iterator] = None

    def _choose_default_sql_stm(self, conf: ConfigTree) -> str:
        url = make_url(conf.get_string(SQLAlchemyExtractor.CONN_STRING))
        if url.drivername.lower() in ['postgresql', 'postgres']:
            return self.DEFAULT_POSTGRES_SQL_STATEMENT
        else:
            return self.DEFAULT_SQL_STATEMENT

    def extract(self) -> Union[TableMetadata, None]:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def get_scope(self) -> str:
        return 'extractor.hive_table_metadata'

    def _get_extract_iter(self) -> Iterator[TableMetadata]:
        """
        Using itertools.groupby and raw level iterator, it groups to table and yields TableMetadata
        :return:
        """
        for key, group in groupby(self._get_raw_extract_iter(), self._get_table_key):
            columns = []

            for row in group:
                last_row = row
                column = None
                if row['is_partition_col'] == 1:
                    # create add a badge to indicate partition column
                    column = ColumnMetadata(row['col_name'], row['col_description'],
                                            row['col_type'], row['col_sort_order'], [PARTITION_BADGE])
                else:
                    column = ColumnMetadata(row['col_name'], row['col_description'],
                                            row['col_type'], row['col_sort_order'])
                columns.append(column)
            is_view = last_row['is_view'] == 1
            yield TableMetadata('hive', self._cluster,
                                last_row['schema'],
                                last_row['name'],
                                last_row['description'],
                                columns,
                                is_view=is_view)

    def _get_raw_extract_iter(self) -> Iterator[Dict[str, Any]]:
        """
        Provides iterator of result row from SQLAlchemy extractor
        :return:
        """
        row = self._alchemy_extractor.extract()
        while row:
            yield row
            row = self._alchemy_extractor.extract()

    def _get_table_key(self, row: Dict[str, Any]) -> Union[TableKey, None]:
        """
        Table key consists of schema and table name
        :param row:
        :return:
        """
        if row:
            return TableKey(schema=row['schema'], table_name=row['name'])

        return None



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import importlib
from typing import (
    Any, Dict, Iterator, Optional,
)

from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.redash.redash_dashboard_utils import (
    RedashPaginatedRestApiQuery, generate_dashboard_description, get_auth_headers, get_text_widgets,
    get_visualization_widgets, sort_widgets,
)
from databuilder.extractor.restapi.rest_api_extractor import REST_API_QUERY, RestAPIExtractor
from databuilder.models.dashboard.dashboard_chart import DashboardChart
from databuilder.models.dashboard.dashboard_last_modified import DashboardLastModifiedTimestamp
from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.dashboard.dashboard_owner import DashboardOwner
from databuilder.models.dashboard.dashboard_query import DashboardQuery
from databuilder.models.dashboard.dashboard_table import DashboardTable
from databuilder.models.table_metadata import TableMetadata
from databuilder.rest_api.base_rest_api_query import EmptyRestApiQuerySeed
from databuilder.rest_api.rest_api_query import RestApiQuery
from databuilder.transformer.base_transformer import ChainedTransformer
from databuilder.transformer.timestamp_string_to_epoch import FIELD_NAME as TS_FIELD_NAME, TimestampStringToEpoch


class TableRelationData:
    """
    This is sort of like a stripped down version of `TableMetadata`.
    It is used as the type returned by the (optional) table parser.
    """

    def __init__(self,
                 database: str,
                 cluster: str,
                 schema: str,
                 name: str) -> None:
        self._data = {'db': database, 'cluster': cluster, 'schema': schema, 'tbl': name}

    @property
    def key(self) -> str:
        return TableMetadata.TABLE_KEY_FORMAT.format(**self._data)


class RedashDashboardExtractor(Extractor):
    """
    An extractor for retrieving dashboards and associated queries
    (and possibly tables) from Redash.

    There are five configuration values:

    - `redash_base_url`: (e.g., `https://redash.example.com`) Base URL for the user-facing
    Redash application
    - `api_base_url`: (e.g., `https://redash.example.com/api`) Base URL for the API
    - `api_key`: Redash API key
    - (optional) `cluster`: A cluster name for this Redash instance (defaults to `prod`)
    - (optional) `table_parser`: A function `(RedashVisualizationWidget) -> List[TableRelationData]`.
    Given a `RedashVisualizationWidget`, this should return a list of potentially related tables
    in Amundsen. Any table returned that exists in Amundsen will be linked to the dashboard.
    Any table that does not exist will be ignored.
    """

    REDASH_BASE_URL_KEY = 'redash_base_url'
    API_BASE_URL_KEY = 'api_base_url'
    API_KEY_KEY = 'api_key'
    CLUSTER_KEY = 'cluster'  # optional config
    TABLE_PARSER_KEY = 'table_parser'  # optional config
    REDASH_VERSION = 'redash_version'  # optional config

    DEFAULT_CLUSTER = 'prod'
    DEFAULT_VERSION = 9

    PRODUCT = 'redash'
    DASHBOARD_GROUP_ID = 'redash'
    DASHBOARD_GROUP_NAME = 'Redash'

    def init(self, conf: ConfigTree) -> None:

        # required configuration
        self._redash_base_url = conf.get_string(RedashDashboardExtractor.REDASH_BASE_URL_KEY)
        self._api_base_url = conf.get_string(RedashDashboardExtractor.API_BASE_URL_KEY)
        self._api_key = conf.get_string(RedashDashboardExtractor.API_KEY_KEY)

        # optional configuration
        self._cluster = conf.get_string(
            RedashDashboardExtractor.CLUSTER_KEY, RedashDashboardExtractor.DEFAULT_CLUSTER
        )
        self._redash_version = conf.get_int(
            RedashDashboardExtractor.REDASH_VERSION, RedashDashboardExtractor.DEFAULT_VERSION
        )

        self._parse_tables = None
        tbl_parser_path = conf.get_string(RedashDashboardExtractor.TABLE_PARSER_KEY)
        if tbl_parser_path:
            module_name, fn_name = tbl_parser_path.rsplit('.', 1)
            mod = importlib.import_module(module_name)
            self._parse_tables = getattr(mod, fn_name)

        self._extractor = self._build_extractor()
        self._transformer = self._build_transformer()
        self._extract_iter: Optional[Iterator[Any]] = None

    def _is_published_dashboard(self, record: Dict[str, Any]) -> bool:
        return not (record['is_archived'] or record['is_draft'])

    def _get_extract_iter(self) -> Iterator[Any]:

        while True:
            record = self._extractor.extract()
            if not record:
                break  # the end.

            record = next(self._transformer.transform(record=record), None)

            if not self._is_published_dashboard(record):
                continue  # filter this one out

            identity_data = {
                'cluster': self._cluster,
                'product': RedashDashboardExtractor.PRODUCT,
                'dashboard_group_id': str(RedashDashboardExtractor.DASHBOARD_GROUP_ID),
                'dashboard_id': str(record['dashboard_id']),
            }

            if self._redash_version >= 9:
                dashboard_url = f'{self._redash_base_url}/dashboards/{record["dashboard_id"]}'
            else:
                dashboard_url = f'{self._redash_base_url}/dashboard/{record["slug"]}'

            dash_data = {
                'dashboard_group':
                    RedashDashboardExtractor.DASHBOARD_GROUP_NAME,
                'dashboard_group_url':
                    self._redash_base_url,
                'dashboard_name':
                    record['dashboard_name'],
                'dashboard_url':
                    dashboard_url,
                'created_timestamp':
                    record['created_timestamp']
            }
            dash_data.update(identity_data)

            widgets = sort_widgets(record['widgets'])
            text_widgets = get_text_widgets(widgets)
            viz_widgets = get_visualization_widgets(widgets)

            # generate a description for this dashboard, since Redash does not have descriptions
            dash_data['description'] = generate_dashboard_description(text_widgets, viz_widgets)

            yield DashboardMetadata(**dash_data)

            last_mod_data = {'last_modified_timestamp': record['last_modified_timestamp']}
            last_mod_data.update(identity_data)

            yield DashboardLastModifiedTimestamp(**last_mod_data)

            owner_data = {'email': record['user']['email']}
            owner_data.update(identity_data)

            yield DashboardOwner(**owner_data)

            table_keys = set()

            for viz in viz_widgets:
                query_data = {
                    'query_id': str(viz.query_id),
                    'query_name': viz.query_name,
                    'url': self._redash_base_url + viz.query_relative_url,
                    'query_text': viz.raw_query
                }

                query_data.update(identity_data)
                yield DashboardQuery(**query_data)

                chart_data = {
                    'query_id': str(viz.query_id),
                    'chart_id': str(viz.visualization_id),
                    'chart_name': viz.visualization_name,
                    'chart_type': viz.visualization_type,
                }
                chart_data.update(identity_data)
                yield DashboardChart(**chart_data)

                # if a table parser is provided, retrieve tables from this viz
                if self._parse_tables:
                    for tbl in self._parse_tables(viz):
                        table_keys.add(tbl.key)

            if len(table_keys) > 0:
                yield DashboardTable(table_ids=list(table_keys), **identity_data)

    def extract(self) -> Any:

        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def _build_restapi_query(self) -> RestApiQuery:

        dashes_query = RedashPaginatedRestApiQuery(
            query_to_join=EmptyRestApiQuerySeed(),
            url=f'{self._api_base_url}/dashboards',
            params=self._get_default_api_query_params(),
            json_path='results[*].[id,name,slug,created_at,updated_at,is_archived,is_draft,user]',
            field_names=[
                'dashboard_id', 'dashboard_name', 'slug', 'created_timestamp',
                'last_modified_timestamp', 'is_archived', 'is_draft', 'user'
            ],
            skip_no_result=True
        )

        if self._redash_version >= 9:
            dashboard_url = f'{self._api_base_url}/dashboards/{{dashboard_id}}'
        else:
            dashboard_url = f'{self._api_base_url}/dashboards/{{slug}}'

        return RestApiQuery(
            query_to_join=dashes_query,
            url=dashboard_url,
            params=self._get_default_api_query_params(),
            json_path='widgets',
            field_names=['widgets'],
            skip_no_result=True
        )

    def _get_default_api_query_params(self) -> Dict[str, Any]:

        return {'headers': get_auth_headers(self._api_key)}

    def _build_extractor(self) -> RestAPIExtractor:

        extractor = RestAPIExtractor()
        rest_api_extractor_conf = ConfigFactory.from_dict({
            REST_API_QUERY: self._build_restapi_query()
        })
        extractor.init(rest_api_extractor_conf)
        return extractor

    def _build_transformer(self) -> ChainedTransformer:

        transformers = []

        # transform timestamps from ISO to unix epoch
        ts_transformer_1 = TimestampStringToEpoch()
        ts_transformer_1.init(ConfigFactory.from_dict({
            TS_FIELD_NAME: 'created_timestamp',
        }))
        transformers.append(ts_transformer_1)

        ts_transformer_2 = TimestampStringToEpoch()
        ts_transformer_2.init(ConfigFactory.from_dict({
            TS_FIELD_NAME: 'last_modified_timestamp',
        }))
        transformers.append(ts_transformer_2)

        return ChainedTransformer(transformers=transformers)

    def get_scope(self) -> str:

        return 'extractor.redash_dashboard'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Iterator, Optional, Union,
)

from amundsen_common.utils.atlas import AtlasCommonParams, AtlasTableTypes
from amundsen_rds.models import RDSModel
from amundsen_rds.models.table import TableSource as RDSTableSource

from databuilder.models.atlas_entity import AtlasEntity
from databuilder.models.atlas_relationship import AtlasRelationship
from databuilder.models.atlas_serializable import AtlasSerializable
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_metadata import TableMetadata
from databuilder.models.table_serializable import TableSerializable
from databuilder.serializers.atlas_serializer import get_entity_attrs
from databuilder.utils.atlas import AtlasRelationshipTypes, AtlasSerializedEntityOperation


class TableSource(GraphSerializable, TableSerializable, AtlasSerializable):
    """
    Hive table source model.
    """
    LABEL = 'Source'
    KEY_FORMAT = '{db}://{cluster}.{schema}/{tbl}/_source'
    SOURCE_TABLE_RELATION_TYPE = 'SOURCE_OF'
    TABLE_SOURCE_RELATION_TYPE = 'SOURCE'

    def __init__(self,
                 db_name: str,
                 schema: str,
                 table_name: str,
                 cluster: str,
                 source: str,
                 source_type: str = 'github',
                 ) -> None:
        self.db = db_name
        self.schema = schema
        self.table = table_name

        self.cluster = cluster if cluster else 'gold'
        # source is the source file location
        self.source = source
        self.source_type = source_type
        self._node_iter = self._create_node_iterator()
        self._relation_iter = self._create_relation_iterator()
        self._record_iter = self._create_record_iterator()
        self._atlas_entity_iterator = self._create_next_atlas_entity()
        self._atlas_relation_iterator = self._create_atlas_relation_iterator()

    def create_next_node(self) -> Optional[GraphNode]:
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def create_next_record(self) -> Union[RDSModel, None]:
        try:
            return next(self._record_iter)
        except StopIteration:
            return None

    def get_source_model_key(self) -> str:
        return TableSource.KEY_FORMAT.format(db=self.db,
                                             cluster=self.cluster,
                                             schema=self.schema,
                                             tbl=self.table)

    def get_metadata_model_key(self) -> str:
        return f'{self.db}://{self.cluster}.{self.schema}/{self.table}'

    def _create_node_iterator(self) -> Iterator[GraphNode]:
        """
        Create a table source node
        :return:
        """
        node = GraphNode(
            key=self.get_source_model_key(),
            label=TableSource.LABEL,
            attributes={
                'source': self.source,
                'source_type': self.source_type
            }
        )
        yield node

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        """
        Create relation map between owner record with original hive table
        :return:
        """
        relationship = GraphRelationship(
            start_label=TableSource.LABEL,
            start_key=self.get_source_model_key(),
            end_label=TableMetadata.TABLE_NODE_LABEL,
            end_key=self.get_metadata_model_key(),
            type=TableSource.SOURCE_TABLE_RELATION_TYPE,
            reverse_type=TableSource.TABLE_SOURCE_RELATION_TYPE,
            attributes={}
        )
        yield relationship

    def _create_record_iterator(self) -> Iterator[RDSModel]:
        record = RDSTableSource(
            rk=self.get_source_model_key(),
            source=self.source,
            source_type=self.source_type,
            table_rk=self.get_metadata_model_key()
        )
        yield record

    def _create_atlas_source_entity(self) -> AtlasEntity:
        attrs_mapping = [
            (AtlasCommonParams.qualified_name, self.get_source_model_key()),
            ('name', self.source),
            ('source_type', self.source_type),
            ('displayName', self.source)
        ]

        entity_attrs = get_entity_attrs(attrs_mapping)

        entity = AtlasEntity(
            typeName=AtlasTableTypes.source,
            operation=AtlasSerializedEntityOperation.CREATE,
            attributes=entity_attrs,
            relationships=None
        )

        return entity

    def create_next_atlas_relation(self) -> Union[AtlasRelationship, None]:
        try:
            return next(self._atlas_relation_iterator)
        except StopIteration:
            return None

    def _create_atlas_relation_iterator(self) -> Iterator[AtlasRelationship]:
        relationship = AtlasRelationship(
            relationshipType=AtlasRelationshipTypes.table_source,
            entityType1=AtlasTableTypes.source,
            entityQualifiedName1=self.get_source_model_key(),
            entityType2=AtlasTableTypes.table,
            entityQualifiedName2=self.get_metadata_model_key(),
            attributes={}
        )

        yield relationship

    def _create_next_atlas_entity(self) -> Iterator[AtlasEntity]:
        yield self._create_atlas_source_entity()

    def create_next_atlas_entity(self) -> Union[AtlasEntity, None]:
        try:
            return next(self._atlas_entity_iterator)
        except StopIteration:
            return None

    def __repr__(self) -> str:
        return f'TableSource({self.db!r}, {self.cluster!r}, {self.schema!r}, {self.table!r}, {self.source!r})'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import importlib
from typing import Any

from pyhocon import ConfigFactory, ConfigTree
from sqlalchemy import create_engine

from databuilder import Scoped
from databuilder.extractor.base_extractor import Extractor


class SQLAlchemyExtractor(Extractor):
    # Config keys
    CONN_STRING = 'conn_string'
    EXTRACT_SQL = 'extract_sql'
    CONNECT_ARGS = 'connect_args'
    """
    An Extractor that extracts records via SQLAlchemy. Database that supports SQLAlchemy can use this extractor
    """

    def init(self, conf: ConfigTree) -> None:
        """
        Establish connections and import data model class if provided
        :param conf:
        """
        self.conf = conf
        self.conn_string = conf.get_string(SQLAlchemyExtractor.CONN_STRING)

        self.connection = self._get_connection()

        self.extract_sql = conf.get_string(SQLAlchemyExtractor.EXTRACT_SQL)

        model_class = conf.get('model_class', None)
        if model_class:
            module_name, class_name = model_class.rsplit(".", 1)
            mod = importlib.import_module(module_name)
            self.model_class = getattr(mod, class_name)
        self._execute_query()

    def close(self) -> None:
        if self.connection is not None:
            self.connection.close()

    def _get_connection(self) -> Any:
        """
        Create a SQLAlchemy connection to Database
        """
        connect_args = {
            k: v
            for k, v in self.conf.get_config(
                self.CONNECT_ARGS, default=ConfigTree()
            ).items()
        }
        engine = create_engine(self.conn_string, connect_args=connect_args)
        conn = engine.connect()
        return conn

    def _execute_query(self) -> None:
        """
        Create an iterator to execute sql.
        """
        if not hasattr(self, 'results'):
            self.results = self.connection.execute(self.extract_sql)

        if hasattr(self, 'model_class'):
            results = [self.model_class(**result)
                       for result in self.results]
        else:
            results = self.results
        self.iter = iter(results)

    def extract(self) -> Any:
        """
        Yield the sql result one at a time.
        convert the result to model if a model_class is provided
        """
        try:
            return next(self.iter)
        except StopIteration:
            return None
        except Exception as e:
            raise e

    def get_scope(self) -> str:
        return 'extractor.sqlalchemy'


def from_surrounding_config(conf: ConfigTree, sql_stmt: str) -> SQLAlchemyExtractor:
    """
    A factory to create SQLAlchemyExtractors that are wrapped by another, specialized
    extractor. This function pulls the config from the wrapping extractor's config, and
    returns a newly configured SQLAlchemyExtractor.
    :param conf: A config tree from which the sqlalchemy config still needs to be taken.
    :param conf: The SQL statement to use for extraction. Expected to be set by the
        wrapping extractor implementation, and not by the config.
    """
    ae = SQLAlchemyExtractor()
    c = Scoped.get_scoped_conf(conf, ae.get_scope()) \
        .with_fallback(ConfigFactory.from_dict({SQLAlchemyExtractor.EXTRACT_SQL: sql_stmt}))
    ae.init(c)
    return ae



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0


import abc
from functools import reduce
from typing import (
    Any, Dict, Iterator, List, Tuple,
)

import requests
from dateutil import parser
from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor.base_extractor import Extractor

type_fields_mapping = List[Tuple[str, str, Any, Any]]


class ApacheSupersetBaseExtractor(Extractor):
    """
    Base class to create extractors pulling any dashboard metadata from Apache Superset.
    """
    APACHE_SUPERSET_PROTOCOL = 'apache_superset_protocol'
    APACHE_SUPERSET_HOST = 'apache_superset_host'
    APACHE_SUPERSET_PORT = 'apache_superset_port'
    APACHE_SUPERSET_SECURITY_SETTINGS_DICT = 'apache_superset_security_settings_dict'
    APACHE_SUPERSET_PAGE_SIZE = 'apache_superset_page_size'
    APACHE_SUPERSET_EXTRACT_PUBLISHED_ONLY = 'apache_superset_extract_published_only'
    APACHE_SUPERSET_SECURITY_PROVIDER = 'apache_superset_security_provider'

    DASHBOARD_GROUP_NAME = 'dashboard_group_name'
    DASHBOARD_GROUP_ID = 'dashboard_group_id'
    DASHBOARD_GROUP_DESCRIPTION = 'dashboard_group_description'

    PRODUCT = 'product'
    CLUSTER = 'cluster'

    DRIVER_TO_DATABASE_MAPPING = 'driver_to_database_mapping'

    DEFAULT_DRIVER_TO_DATABASE_MAPPING = {
        'postgresql': 'postgres',
        'mysql+pymysql': 'mysql'
    }

    DATABASE_TO_CLUSTER_MAPPING = 'database_to_cluster_mapping'  # map superset dbs to preferred clusters

    DEFAULT_CONFIG = ConfigFactory.from_dict({
        APACHE_SUPERSET_PROTOCOL: 'http',
        APACHE_SUPERSET_HOST: 'localhost',
        APACHE_SUPERSET_PORT: '8088',
        APACHE_SUPERSET_PAGE_SIZE: 20,
        APACHE_SUPERSET_EXTRACT_PUBLISHED_ONLY: False,
        PRODUCT: 'superset',
        DASHBOARD_GROUP_DESCRIPTION: '',
        DRIVER_TO_DATABASE_MAPPING: DEFAULT_DRIVER_TO_DATABASE_MAPPING,
        DATABASE_TO_CLUSTER_MAPPING: {}
    })

    def init(self, conf: ConfigTree) -> None:
        self.conf = conf.with_fallback(ApacheSupersetBaseExtractor.DEFAULT_CONFIG)
        self._extract_iter = self._get_extract_iter()

        self.authenticate()

    def get_scope(self) -> str:
        return 'extractor.apache_superset'

    def extract(self) -> Any:
        try:
            result = next(self._extract_iter)

            return result
        except StopIteration:
            return None

    def authenticate(self) -> None:
        security_settings = dict(self.conf.get(ApacheSupersetBaseExtractor.APACHE_SUPERSET_SECURITY_SETTINGS_DICT))

        token = requests.post(self.build_full_url('api/v1/security/login'),
                              json=security_settings)

        self.token = token.json()['access_token']

    def build_full_url(self, endpoint: str) -> str:
        return f'{self.base_url}/{endpoint}'

    def execute_query(self, url: str, params: dict = {}) -> Dict:
        try:
            data = requests.get(url, params=params, headers={'Authorization': f'Bearer {self.token}'})

            if data.status_code == 401:
                self.authenticate()

                return self.execute_query(url, params)
            else:
                return data.json()
        except Exception:
            return {}

    @property
    def base_url(self) -> str:
        _protocol = self.conf.get(ApacheSupersetBaseExtractor.APACHE_SUPERSET_PROTOCOL)
        _host = self.conf.get(ApacheSupersetBaseExtractor.APACHE_SUPERSET_HOST)
        _port = self.conf.get(ApacheSupersetBaseExtractor.APACHE_SUPERSET_PORT)

        return f'{_protocol}://{_host}:{_port}'

    @property
    def page_size(self) -> int:
        return self.conf.get_int(ApacheSupersetBaseExtractor.APACHE_SUPERSET_PAGE_SIZE)

    @property
    def product(self) -> str:
        return self.conf.get(ApacheSupersetBaseExtractor.PRODUCT)

    @property
    def cluster(self) -> str:
        return self.conf.get(ApacheSupersetBaseExtractor.CLUSTER)

    @property
    def extract_published_only(self) -> bool:
        return self.conf.get(ApacheSupersetBaseExtractor.APACHE_SUPERSET_EXTRACT_PUBLISHED_ONLY)

    @property
    def common_params(self) -> Dict[str, str]:
        return dict(dashboard_group=self.conf.get(ApacheSupersetBaseExtractor.DASHBOARD_GROUP_NAME),
                    dashboard_group_id=self.conf.get(ApacheSupersetBaseExtractor.DASHBOARD_GROUP_ID),
                    dashboard_group_url=self.base_url,
                    dashboard_group_description=self.conf.get(ApacheSupersetBaseExtractor.DASHBOARD_GROUP_DESCRIPTION),
                    product=self.product,
                    cluster=self.cluster)

    @staticmethod
    def parse_date(string_date: str) -> int:
        try:
            date_parsed = parser.parse(string_date)

            # date returned by superset api does not contain timezone so to be timezone safe we need to assume it's utc
            if not date_parsed.tzname():
                return ApacheSupersetBaseExtractor.parse_date(f'{string_date}+0000')

            return int(date_parsed.timestamp())
        except Exception:
            return 0

    @staticmethod
    def get_nested_field(input_dict: Dict, field: str) -> Any:
        return reduce(lambda x, y: x.get(y, dict()), field.split('.'), input_dict)

    @staticmethod
    def map_fields(data: Dict, mappings: type_fields_mapping) -> Dict:
        result = dict()

        for mapping in mappings:
            target_field, source_field, transform, default_value = mapping
            value = ApacheSupersetBaseExtractor.get_nested_field(data, source_field)

            if transform:
                value = transform(value)

            result[target_field] = value or default_value

        return result

    @abc.abstractmethod
    def _get_extract_iter(self) -> Iterator[Any]:
        pass

    def _get_resource_ids(self, resource: str) -> List[str]:
        i = 0
        result = []

        while True:
            url = self.build_full_url(f'api/v1/{resource}?q=(page_size:{self.page_size},page:{i},order_direction:desc)')

            data = self.execute_query(url)

            ids = data.get('ids', [])

            if ids:
                result += ids
                i += 1
            else:
                break

        return result



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import json
from typing import (
    Any, Dict, Tuple,
)

import dateutil.parser
from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_stats import TableColumnStats


class PandasProfilingColumnStatsExtractor(Extractor):
    FILE_PATH = 'file_path'
    DATABASE_NAME = 'database_name'
    TABLE_NAME = 'table_name'
    SCHEMA_NAME = 'schema_name'
    CLUSTER_NAME = 'cluster_name'

    # if you wish to collect only selected set of metrics configure stat_mappings option of the extractor providing
    # similar dictionary but containing only keys of metrics you wish to collect.
    # For example - if you want only min and max value of a column, provide extractor with configuration option:
    # PandasProfilingColumnStatsExtractor.STAT_MAPPINGS = {'max': ('Maximum', float), 'min': ('Minimum', float)}
    STAT_MAPPINGS = 'stat_mappings'

    # - key - raw name of the stat in pandas-profiling. Value - tuple of stat spec.
    # - first value of the tuple - full name of the stat
    # - second value of the tuple - function modifying the stat (by default we just do type casting)
    DEFAULT_STAT_MAPPINGS = {
        '25%': ('Quantile 25%', float),
        '5%': ('Quantile 5%', float),
        '50%': ('Quantile 50%', float),
        '75%': ('Quantile 75%', float),
        '95%': ('Quantile 95%', float),
        'chi_squared': ('Chi squared', lambda x: float(x.get('statistic'))),
        'count': ('Count', int),
        'is_unique': ('Unique', bool),
        'kurtosis': ('Kurtosis', float),
        'max': ('Maximum', str),
        'max_length': ('Maximum length', int),
        'mean': ('Mean', float),
        'mean_length': ('Mean length', int),
        'median_length': ('Median length', int),
        'min': ('Minimum', str),
        'min_length': ('Minimum length', int),
        'monotonic': ('Monotonic', bool),
        'n_characters': ('Characters', int),
        'n_characters_distinct': ('Distinct characters', int),
        'n_distinct': ('Distinct values', int),
        'n_infinite': ('Infinite values', int),
        'n_missing': ('Missing values', int),
        'n_negative': ('Negative values', int),
        'n_unique': ('Unique values', int),
        'n_zeros': ('Zeros', int),
        'p_distinct': ('Distinct values %', lambda x: float(x * 100)),
        'p_infinite': ('Infinite values %', lambda x: float(x * 100)),
        'p_missing': ('Missing values %', lambda x: float(x * 100)),
        'p_negative': ('Negative values %', lambda x: float(x * 100)),
        'p_unique': ('Unique values %', lambda x: float(x * 100)),
        'p_zeros': ('Zeros %', lambda x: float(x * 100)),
        'range': ('Range', str),
        'skewness': ('Skewness', float),
        'std': ('Std. deviation', float),
        'sum': ('Sum', float),
        'variance': ('Variance', float)
        # Stats available in pandas-profiling but are not collected by default and require custom, conscious config..
        # 'block_alias_char_counts': ('',),
        # 'block_alias_counts': ('',),
        # 'block_alias_values': ('',),
        # 'category_alias_char_counts': ('',),
        # 'category_alias_counts': ('',),
        # 'category_alias_values': ('',),
        # 'character_counts': ('',),
        # 'cv': ('',),
        # 'first_rows': ('',),
        # 'hashable': ('',),
        # 'histogram': ('',),
        # 'histogram_frequencies': ('',),
        # 'histogram_length': ('',),
        # 'iqr': ('',),
        # 'length': ('',),
        # 'mad': ('',),
        # 'memory_size': ('',),
        # 'monotonic_decrease': ('Monotonic decrease', bool),
        # 'monotonic_decrease_strict': ('Strict monotonic decrease', bool),
        # 'monotonic_increase': ('Monotonic increase', bool),
        # 'monotonic_increase_strict': ('Strict monotonic increase', bool),
        # 'n': ('',),
        # 'n_block_alias': ('',),
        # 'n_category': ('Categories', int),
        # 'n_scripts': ('',),
        # 'ordering': ('',),
        # 'script_char_counts': ('',),
        # 'script_counts': ('',),
        # 'value_counts_index_sorted': ('',),
        # 'value_counts_without_nan': ('',),
        # 'word_counts': ('',),
        # 'type': ('Type', str)
    }

    PRECISION = 'precision'

    DEFAULT_CONFIG = ConfigFactory.from_dict({STAT_MAPPINGS: DEFAULT_STAT_MAPPINGS, PRECISION: 3})

    def get_scope(self) -> str:
        return 'extractor.pandas_profiling'

    def init(self, conf: ConfigTree) -> None:
        self.conf = conf.with_fallback(PandasProfilingColumnStatsExtractor.DEFAULT_CONFIG)

        self._extract_iter = self._get_extract_iter()

    def extract(self) -> Any:
        try:
            result = next(self._extract_iter)

            return result
        except StopIteration:
            return None

    def _get_extract_iter(self) -> Any:
        report = self._load_report()

        variables = report.get('variables', dict())
        report_time = self.parse_date(report.get('analysis', dict()).get('date_start'))

        for column_name, column_stats in variables.items():
            for _stat_name, stat_value in column_stats.items():
                stat_spec = self.stat_mappings.get(_stat_name)

                if stat_spec:
                    stat_name, stat_modifier = stat_spec

                    if isinstance(stat_value, float):
                        stat_value = self.round_value(stat_value)

                    stat = TableColumnStats(table_name=self.table_name, col_name=column_name, stat_name=stat_name,
                                            stat_val=stat_modifier(stat_value), start_epoch=report_time, end_epoch='0',
                                            db=self.database_name, cluster=self.cluster_name, schema=self.schema_name)

                    yield stat

    def _load_report(self) -> Dict[str, Any]:
        path = self.conf.get(PandasProfilingColumnStatsExtractor.FILE_PATH)

        try:
            with open(path, 'r') as f:
                _data = f.read()

            data = json.loads(_data)

            return data
        except Exception:
            return {}

    @staticmethod
    def parse_date(string_date: str) -> str:
        try:
            date_parsed = dateutil.parser.parse(string_date)

            # date from pandas-profiling doesn't contain timezone so to be timezone safe we need to assume it's utc
            if not date_parsed.tzname():
                return PandasProfilingColumnStatsExtractor.parse_date(f'{string_date}+0000')

            return str(int(date_parsed.timestamp()))
        except Exception:
            return '0'

    def round_value(self, value: float) -> float:
        return round(value, self.conf.get(PandasProfilingColumnStatsExtractor.PRECISION))

    @property
    def stat_mappings(self) -> Dict[str, Tuple[str, Any]]:
        return dict(self.conf.get(PandasProfilingColumnStatsExtractor.STAT_MAPPINGS))

    @property
    def cluster_name(self) -> str:
        return self.conf.get(PandasProfilingColumnStatsExtractor.CLUSTER_NAME)

    @property
    def database_name(self) -> str:
        return self.conf.get(PandasProfilingColumnStatsExtractor.DATABASE_NAME)

    @property
    def schema_name(self) -> str:
        return self.conf.get(PandasProfilingColumnStatsExtractor.SCHEMA_NAME)

    @property
    def table_name(self) -> str:
        return self.conf.get(PandasProfilingColumnStatsExtractor.TABLE_NAME)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

ORGANIZATION = 'organization'
MODE_ACCESS_TOKEN = 'mode_user_token'
MODE_PASSWORD_TOKEN = 'mode_password_token'

# this token is needed to access batch discover endpoint
# e.g https://mode.com/developer/discovery-api/introduction/
MODE_BEARER_TOKEN = 'mode_bearer_token'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0


import logging
from collections import namedtuple
from itertools import groupby
from typing import (
    Any, Dict, Iterator, Union,
)

from pyhocon import ConfigFactory, ConfigTree
from pyodbc import connect

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata

TableKey = namedtuple('TableKey', ['schema', 'table_name'])

LOGGER = logging.getLogger(__name__)


class DremioMetadataExtractor(Extractor):
    '''
    Extracts Dremio table and column metadata from underlying INFORMATION_SCHEMA table

    Requirements:
        pyodbc & Dremio driver
    '''

    SQL_STATEMENT = '''
    SELECT
      nested_1.COLUMN_NAME AS col_name,
      CAST(NULL AS VARCHAR) AS col_description,
      nested_1.DATA_TYPE AS col_type,
      nested_1.ORDINAL_POSITION AS col_sort_order,
      nested_1.TABLE_CATALOG AS database,
      '{cluster}' AS cluster,
      nested_1.TABLE_SCHEMA AS schema,
      nested_1.TABLE_NAME AS name,
      CAST(NULL AS VARCHAR) AS description,
      CASE WHEN nested_0.TABLE_TYPE='VIEW' THEN TRUE ELSE FALSE END AS is_view
    FROM (
      SELECT TABLE_CATALOG, TABLE_SCHEMA, TABLE_NAME, TABLE_TYPE
      FROM INFORMATION_SCHEMA."TABLES"
    ) nested_0
    RIGHT JOIN (
      SELECT TABLE_CATALOG, TABLE_SCHEMA, TABLE_NAME, COLUMN_NAME, DATA_TYPE, ORDINAL_POSITION
      FROM INFORMATION_SCHEMA."COLUMNS"
    ) nested_1 ON nested_0.TABLE_NAME = nested_1.TABLE_NAME
      AND nested_0.TABLE_SCHEMA = nested_1.TABLE_SCHEMA
      AND nested_0.TABLE_CATALOG = nested_1.TABLE_CATALOG
    {where_stmt}
    '''

    # Config keys
    DREMIO_USER_KEY = 'user_key'
    DREMIO_PASSWORD_KEY = 'password_key'
    DREMIO_HOST_KEY = 'host_key'
    DREMIO_PORT_KEY = 'port_key'
    DREMIO_DRIVER_KEY = 'driver_key'
    DREMIO_CLUSTER_KEY = 'cluster_key'
    DREMIO_EXCLUDE_SYS_TABLES_KEY = 'exclude_system_tables'
    DREMIO_EXCLUDE_PDS_TABLES_KEY = 'exclude_pds_tables'

    # Default values
    DEFAULT_AUTH_USER = 'dremio_auth_user'
    DEFAULT_AUTH_PW = 'dremio_auth_pw'
    DEFAULT_HOST = 'localhost'
    DEFAULT_PORT = '31010'
    DEFAULT_DRIVER = 'DSN=Dremio Connector'
    DEFAULT_CLUSTER_NAME = 'Production'
    DEFAULT_EXCLUDE_SYS_TABLES = True
    DEFAULT_EXCLUDE_PDS_TABLES = False

    # Default config
    DEFAULT_CONFIG = ConfigFactory.from_dict({
        DREMIO_USER_KEY: DEFAULT_AUTH_USER,
        DREMIO_PASSWORD_KEY: DEFAULT_AUTH_PW,
        DREMIO_HOST_KEY: DEFAULT_HOST,
        DREMIO_PORT_KEY: DEFAULT_PORT,
        DREMIO_DRIVER_KEY: DEFAULT_DRIVER,
        DREMIO_CLUSTER_KEY: DEFAULT_CLUSTER_NAME,
        DREMIO_EXCLUDE_SYS_TABLES_KEY: DEFAULT_EXCLUDE_SYS_TABLES,
        DREMIO_EXCLUDE_PDS_TABLES_KEY: DEFAULT_EXCLUDE_PDS_TABLES
    })

    def init(self, conf: ConfigTree) -> None:
        conf = conf.with_fallback(DremioMetadataExtractor.DEFAULT_CONFIG)

        exclude_sys_tables = conf.get_bool(DremioMetadataExtractor.DREMIO_EXCLUDE_SYS_TABLES_KEY)
        exclude_pds_tables = conf.get_bool(DremioMetadataExtractor.DREMIO_EXCLUDE_PDS_TABLES_KEY)
        if exclude_sys_tables and exclude_pds_tables:
            where_stmt = ("WHERE nested_0.TABLE_TYPE != 'SYSTEM_TABLE' AND "
                          "nested_0.TABLE_TYPE != 'TABLE';")
        elif exclude_sys_tables:
            where_stmt = "WHERE nested_0.TABLE_TYPE != 'SYSTEM_TABLE';"
        elif exclude_pds_tables:
            where_stmt = "WHERE nested_0.TABLE_TYPE != 'TABLE';"
        else:
            where_stmt = ';'

        self._cluster = conf.get_string(DremioMetadataExtractor.DREMIO_CLUSTER_KEY)

        self._cluster = conf.get_string(DremioMetadataExtractor.DREMIO_CLUSTER_KEY)

        self.sql_stmt = DremioMetadataExtractor.SQL_STATEMENT.format(
            cluster=self._cluster,
            where_stmt=where_stmt
        )

        LOGGER.info('SQL for Dremio metadata: %s', self.sql_stmt)

        self._pyodbc_cursor = connect(
            conf.get_string(DremioMetadataExtractor.DREMIO_DRIVER_KEY),
            uid=conf.get_string(DremioMetadataExtractor.DREMIO_USER_KEY),
            pwd=conf.get_string(DremioMetadataExtractor.DREMIO_PASSWORD_KEY),
            host=conf.get_string(DremioMetadataExtractor.DREMIO_HOST_KEY),
            port=conf.get_string(DremioMetadataExtractor.DREMIO_PORT_KEY),
            autocommit=True).cursor()

        self._extract_iter: Union[None, Iterator] = None

    def extract(self) -> Union[TableMetadata, None]:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def get_scope(self) -> str:
        return 'extractor.dremio'

    def _get_extract_iter(self) -> Iterator[TableMetadata]:
        '''
        Using itertools.groupby and raw level iterator, it groups to table and yields TableMetadata
        :return:
        '''
        for _, group in groupby(self._get_raw_extract_iter(), self._get_table_key):
            columns = []

            for row in group:
                last_row = row
                columns.append(ColumnMetadata(
                               row['col_name'],
                               row['col_description'],
                               row['col_type'],
                               row['col_sort_order'])
                               )

            yield TableMetadata(last_row['database'],
                                last_row['cluster'],
                                last_row['schema'],
                                last_row['name'],
                                last_row['description'],
                                columns,
                                last_row['is_view'] == 'true')

    def _get_raw_extract_iter(self) -> Iterator[Dict[str, Any]]:
        '''
        Provides iterator of result row from SQLAlchemy extractor
        :return:
        '''

        for row in self._pyodbc_cursor.execute(self.sql_stmt):
            yield dict(zip([c[0] for c in self._pyodbc_cursor.description], row))

    def _get_table_key(self, row: Dict[str, Any]) -> Union[TableKey, None]:
        '''
        Table key consists of schema and table name
        :param row:
        :return:
        '''
        if row:
            return TableKey(schema=row['schema'], table_name=row['name'])

        return None



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Iterator, Optional, Union,
)

from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.query.query import QueryMetadata


class QueryExecutionsMetadata(GraphSerializable):
    """
    The Amundsen Query Executions model represents an aggregation for the number
    of times a query was executed within a given time window.

    Query executions are aggregated to time window to enable easily adding and
    dropping new query execution aggregations without having to maintain
    all instances that a query was executed in the database. Query Executions only
    contain a start time and a window duration, although the window duration is
    only used for informational purposes. Amundsen does not apply validation that
    query executions do not overlap, therefore, it is important that any implementation
    of a query execution is able to deterministically retrieve non-overlapping queries
    between data builder runs.
    """
    NODE_LABEL = 'Execution'
    KEY_FORMAT = '{query_key}-{start_time}'

    # Relation between entity and query
    QUERY_EXECUTION_RELATION_TYPE = 'HAS_EXECUTION'
    INVERSE_QUERY_EXECUTION_RELATION_TYPE = 'EXECUTION_OF'

    # Execution window ENUMs
    EXECUTION_WINDOW_HOURLY = 'hourly'
    EXECUTION_WINDOW_DAILY = 'daily'
    EXECUTION_WINDOW_WEEKLY = 'weekly'

    # Attributes
    START_TIME = 'start_time'
    EXECUTION_COUNT = 'execution_count'
    WINDOW_DURATION = 'window_duration'

    def __init__(self,
                 query_metadata: QueryMetadata,
                 start_time: int,
                 execution_count: int,
                 window_duration: str = EXECUTION_WINDOW_DAILY,     # Purely for descriptive purposes
                 yield_relation_nodes: bool = False
                 ):
        """
        :param query_metadata: The Query metadata object that this execution belongs to
        :param start_time: The time the query execution window started. This should
            consistently be supplied as either seconds or milliseconds since epoch.
        :param execution_count: The count of the number of times this query was executed
            within the window
        :param window_duration: A description of the window duration, e.g. daily, hourly
        :param yield_relation_nodes: A boolean, indicating whether or not the query
            associated to this execution should have it's node created if it does not
            already exist.
        """
        self.query_metadata = query_metadata
        self.start_time = start_time
        self.execution_count = execution_count
        self.window_duration = window_duration
        self.yield_relation_nodes = yield_relation_nodes
        self._node_iter = self._create_next_node()
        self._relation_iter = self._create_relation_iterator()

    def __repr__(self) -> str:
        return (
            f'QueryExecutionsMetadata(Query: {self.query_metadata.get_key_self()}, Start Time: {self.start_time}, '
            f'Window Duration: {self.window_duration}, Count: {self.execution_count})'
        )

    def create_next_node(self) -> Optional[GraphNode]:
        # return the string representation of the data
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    @staticmethod
    def get_key(query_key: str, start_time: Union[str, int]) -> str:
        return QueryExecutionsMetadata.KEY_FORMAT.format(query_key=query_key, start_time=start_time)

    def get_key_self(self) -> str:
        return QueryExecutionsMetadata.get_key(query_key=self.query_metadata.get_key_self(), start_time=self.start_time)

    def get_query_relations(self) -> Iterator[GraphRelationship]:
        yield GraphRelationship(
            start_label=QueryMetadata.NODE_LABEL,
            end_label=self.NODE_LABEL,
            start_key=self.query_metadata.get_key_self(),
            end_key=self.get_key_self(),
            type=self.QUERY_EXECUTION_RELATION_TYPE,
            reverse_type=self.INVERSE_QUERY_EXECUTION_RELATION_TYPE,
            attributes={}
        )

    def _create_next_node(self) -> Iterator[GraphNode]:
        """
        Create query nodes
        :return:
        """
        # TODO: Should query metadata yiled tables as well? Otherwise if a table does not exist
        # before this script is ran then the query relatoinship will not get created.
        # Ideally this relationship wouldn't be "lost" but created once the table is created as well.
        yield GraphNode(
            key=self.get_key_self(),
            label=self.NODE_LABEL,
            attributes={
                self.START_TIME: self.start_time,
                self.EXECUTION_COUNT: self.execution_count,
                self.WINDOW_DURATION: self.window_duration,
            }
        )
        if self.yield_relation_nodes and self.query_metadata:
            for query_item in self.query_metadata._create_next_node():
                yield query_item

    def _create_relation_iterator(self) -> Iterator[GraphRelationship]:
        relations = self.get_query_relations()
        for relation in relations:
            yield relation

        if self.yield_relation_nodes and self.query_metadata:
            for query_rel in self.query_metadata._create_relation_iterator():
                yield query_rel



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    Any, Dict, Iterator, Optional,
)

from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.extractor.dashboard.databricks_sql.databricks_sql_dashboard_utils import (
    DatabricksSQLPaginatedRestApiQuery, generate_dashboard_description, get_text_widgets, get_visualization_widgets,
    sort_widgets,
)
from databuilder.extractor.restapi.rest_api_extractor import REST_API_QUERY, RestAPIExtractor
from databuilder.models.dashboard.dashboard_chart import DashboardChart
from databuilder.models.dashboard.dashboard_last_modified import DashboardLastModifiedTimestamp
from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.models.dashboard.dashboard_owner import DashboardOwner
from databuilder.models.dashboard.dashboard_query import DashboardQuery
from databuilder.rest_api.base_rest_api_query import EmptyRestApiQuerySeed
from databuilder.rest_api.rest_api_query import RestApiQuery
from databuilder.transformer.base_transformer import ChainedTransformer
from databuilder.transformer.timestamp_string_to_epoch import FIELD_NAME as TS_FIELD_NAME, TimestampStringToEpoch


class DatabricksSQLDashboardExtractor(Extractor):
    """
    An extractor for retrieving dashboards, queries, and visualizations
    from Databricks SQL (https://databricks.com/product/databricks-sql)
    """

    DATABRICKS_HOST_KEY = "databricks_host"
    DATABRICKS_API_TOKEN_KEY = "databricks_api_token"

    PRODUCT = "databricks-sql"
    DASHBOARD_GROUP_ID = "databricks-sql"
    DASHBOARD_GROUP_NAME = "Databricks SQL"

    def init(self, conf: ConfigTree) -> None:
        # Required configuration
        self._databricks_host = conf.get_string(
            DatabricksSQLDashboardExtractor.DATABRICKS_HOST_KEY
        )
        self._databricks_api_token = conf.get_string(
            DatabricksSQLDashboardExtractor.DATABRICKS_API_TOKEN_KEY
        )

        # NOTE: The dashboards api is currently in preview. When it gets moved out of preview
        # this will break and it will need to be changed
        self._databricks_sql_dashboards_api_base = (
            f"{self._databricks_host}/api/2.0/preview/sql/dashboards"
        )

        self._extractor = self._build_extractor()
        self._transformer = self._build_transformer()
        self._extract_iter: Optional[Iterator[Any]] = None

    def _get_databrick_request_headers(self) -> Dict[str, str]:
        return {
            "Authorization": f"Bearer {self._databricks_api_token}",
        }

    def _get_extract_iter(self) -> Iterator[Any]:
        while True:
            record = self._extractor.extract()
            if not record:
                break

            record = next(self._transformer.transform(record=record), None)
            dashboard_identity_data = {
                "dashboard_group_id": DatabricksSQLDashboardExtractor.DASHBOARD_GROUP_ID,
                "dashboard_id": record["dashboard_id"],
                "product": "databricks-sql",
            }

            dashboard_data = {
                "dashboard_group": DatabricksSQLDashboardExtractor.DASHBOARD_GROUP_NAME,
                "dashboard_name": record["dashboard_name"],
                "dashboard_url": f"{self._databricks_host}/sql/dashboards/{record['dashboard_id']}",
                "dashboard_group_url": self._databricks_host,
                "created_timestamp": record["created_timestamp"],
                "tags": record["tags"],
            }

            dashboard_owner_data = {"email": record["user"]["email"]}
            dashboard_owner_data.update(dashboard_identity_data)
            yield DashboardOwner(**dashboard_owner_data)

            dashboard_last_modified_data = {
                "last_modified_timestamp": record["last_modified_timestamp"],
            }
            dashboard_last_modified_data.update(dashboard_identity_data)
            yield DashboardLastModifiedTimestamp(**dashboard_last_modified_data)

            if "widgets" in record:
                widgets = sort_widgets(record["widgets"])
                text_widgets = get_text_widgets(widgets)
                viz_widgets = get_visualization_widgets(widgets)
                dashboard_data["description"] = generate_dashboard_description(
                    text_widgets, viz_widgets
                )

                for viz in viz_widgets:
                    dashboard_query_data = {
                        "query_id": str(viz.query_id),
                        "query_name": viz.query_name,
                        "url": self._databricks_host + viz.query_relative_url,
                        "query_text": viz.raw_query,
                    }
                    dashboard_query_data.update(dashboard_identity_data)
                    yield DashboardQuery(**dashboard_query_data)

                    dashboard_chart_data = {
                        "query_id": str(viz.query_id),
                        "chart_id": str(viz.visualization_id),
                        "chart_name": viz.visualization_name,
                        "chart_type": viz.visualization_type,
                    }
                    dashboard_chart_data.update(dashboard_identity_data)
                    yield DashboardChart(**dashboard_chart_data)

            dashboard_data.update(dashboard_identity_data)
            yield DashboardMetadata(**dashboard_data)

    def extract(self) -> Any:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def _build_extractor(self) -> RestAPIExtractor:
        extractor = RestAPIExtractor()
        rest_api_extractor_conf = ConfigFactory.from_dict(
            {REST_API_QUERY: self._build_restapi_query()}
        )
        extractor.init(rest_api_extractor_conf)
        return extractor

    def _build_transformer(self) -> ChainedTransformer:
        transformers = []
        # transform timestamps from ISO to unix epoch
        ts_transformer_1 = TimestampStringToEpoch()
        ts_transformer_1.init(
            ConfigFactory.from_dict({TS_FIELD_NAME: "created_timestamp"})
        )
        transformers.append(ts_transformer_1)

        ts_transformer_2 = TimestampStringToEpoch()
        ts_transformer_2.init(
            ConfigFactory.from_dict({TS_FIELD_NAME: "last_modified_timestamp"})
        )
        transformers.append(ts_transformer_2)

        return ChainedTransformer(transformers=transformers)

    def _build_restapi_query(self) -> RestApiQuery:
        databricks_sql_dashboard_query = DatabricksSQLPaginatedRestApiQuery(
            query_to_join=EmptyRestApiQuerySeed(),
            url=self._databricks_sql_dashboards_api_base,
            params={"headers": self._get_databrick_request_headers()},
            json_path="results[*].[id,name,tags,updated_at,created_at,user]",
            field_names=[
                "dashboard_id",
                "dashboard_name",
                "tags",
                "last_modified_timestamp",
                "created_timestamp",
                "user",
            ],
            skip_no_results=True,
        )

        return RestApiQuery(
            query_to_join=databricks_sql_dashboard_query,
            url=f"{self._databricks_sql_dashboards_api_base}/{{dashboard_id}}",
            params={"headers": self._get_databrick_request_headers()},
            json_path="widgets",
            field_names=["widgets"],
            skip_no_result=True,
        )

    def get_scope(self) -> str:
        return "extractor.databricks_sql_extractor"



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import json
import logging
from typing import (
    Any, Dict, Iterator,
)

from pyhocon import ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_lineage import TableLineage

LOGGER = logging.getLogger(__name__)


class OpenLineageTableLineageExtractor(Extractor):
    # Config keys
    TABLE_LINEAGE_FILE_LOCATION = 'table_lineage_file_location'
    CLUSTER_NAME = 'cluster_name'
    OL_DATASET_NAMESPACE_OVERRIDE = 'namespace_override'
    # Openlineage values key's, which will be used to extract data from an OpenLineage event
    OL_INPUTS_KEY = 'inputs_key'
    OL_OUTPUTS_KEY = 'outputs_key'
    OL_DATASET_NAMESPACE_KEY = 'namespace_key'
    OL_DATASET_DATABASE_KEY = 'database_key'
    OL_DATASET_NAME_KEY = 'dataset_name_key'

    """
    An Extractor that creates Table Lineage between two tables based on OpenLineage event
    """

    def init(self, conf: ConfigTree) -> None:
        """
        :param conf:
        """
        self.conf = conf
        self.table_lineage_file_location = conf.get_string(OpenLineageTableLineageExtractor.TABLE_LINEAGE_FILE_LOCATION)
        self.cluster_name = conf.get_string(OpenLineageTableLineageExtractor.CLUSTER_NAME)
        self.ol_inputs_key = conf.get_string(OpenLineageTableLineageExtractor.OL_INPUTS_KEY, default='inputs')
        self.ol_outputs_key = conf.get_string(OpenLineageTableLineageExtractor.OL_OUTPUTS_KEY, default='outputs')
        self.ol_namespace_key = conf.get_string(
            OpenLineageTableLineageExtractor.OL_DATASET_NAMESPACE_KEY, default='namespace')
        self.ol_database_key = conf.get_string(
            OpenLineageTableLineageExtractor.OL_DATASET_DATABASE_KEY, default='database')
        self.ol_dataset_name_key = conf.get_string(
            OpenLineageTableLineageExtractor.OL_DATASET_NAME_KEY, default='name')
        self.ol_namespace_override = conf.get_string(
            OpenLineageTableLineageExtractor.OL_DATASET_NAMESPACE_OVERRIDE, default=None)
        self._load_openlineage_event()

    def _extract_dataset_info(self, openlineage_event: Any) -> Iterator[Dict]:
        """
        Yield input/output dict in form of amundsen table keys
        """

        for event in openlineage_event:
            try:
                in_and_outs = ((inputs, outputs)
                               for inputs in event[self.ol_inputs_key]
                               for outputs in event[self.ol_outputs_key])
                for row in in_and_outs:
                    yield {'input': self._amundsen_dataset_key(row[0]),
                           'output': self._amundsen_dataset_key(row[1])}
            except KeyError:
                LOGGER.error(f'Cannot extract valid input or output from Openlineage event \n {event} ')

    def _amundsen_dataset_key(self, dataset: Dict) -> str:
        """
        Generation of amundsen dataset key with optional namespace overriding.
        Amundsen dataset key format: <namespace>://<cluster_name>.<database>/<table>.
        If dataset name is represented in path form ie. ( /warehouse/database/table )
        only last part of such path will be extracted as dataset name
        """
        namespace = self.ol_namespace_override if self.ol_namespace_override else dataset[self.ol_namespace_key]
        return f'{namespace}://{self.cluster_name}.{dataset[self.ol_database_key]}' \
               f'/{dataset[self.ol_dataset_name_key].split("/")[-1]}'

    def _load_openlineage_event(self) -> Any:

        self.input_file = open(self.table_lineage_file_location, 'r')

        lineage_event = (json.loads(line) for line in self.input_file)

        table_lineage = (TableLineage(table_key=lineage['input'],
                                      downstream_deps=[lineage['output']])

                         for lineage in self._extract_dataset_info(lineage_event))
        self._iter = table_lineage

    def extract(self) -> Any:
        """
        Yield the csv result one at a time.
        convert the result to model if a model_class is provided
        """
        try:
            return next(self._iter)
        except StopIteration:
            self.input_file.close()
            return None
        except Exception as e:
            raise e

    def get_scope(self) -> str:
        return 'extractor.openlineage_tablelineage'



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import abc
import logging
from typing import (
    Any, Dict, Iterator, List, Optional, Union,
)

from databuilder.models.badge import Badge, BadgeMetadata
from databuilder.models.description_metadata import DescriptionMetadata
from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import GraphSerializable
from databuilder.models.table_metadata import ColumnMetadata, _format_as_list

LOGGER = logging.getLogger(__name__)


class TypeMetadata(abc.ABC, GraphSerializable):
    NODE_LABEL = 'Type_Metadata'
    COL_TM_RELATION_TYPE = 'TYPE_METADATA'
    TM_COL_RELATION_TYPE = 'TYPE_METADATA_OF'
    SUBTYPE_RELATION_TYPE = 'SUBTYPE'
    INVERSE_SUBTYPE_RELATION_TYPE = 'SUBTYPE_OF'
    KIND = 'kind'
    NAME = 'name'
    DATA_TYPE = 'data_type'
    SORT_ORDER = 'sort_order'

    @abc.abstractmethod
    def __init__(self,
                 name: str,
                 parent: Union[ColumnMetadata, 'TypeMetadata'],
                 type_str: str,
                 sort_order: Optional[int] = None) -> None:
        self.name = name
        self.parent = parent
        self.type_str = type_str
        # Sort order among TypeMetadata objects with the same parent
        self.sort_order = sort_order

        self._description: Optional[DescriptionMetadata] = None
        self._badges: Optional[List[Badge]] = None

        self._node_iter = self.create_node_iterator()
        self._relation_iter = self.create_relation_iterator()

    def get_description(self) -> Optional[DescriptionMetadata]:
        return self._description

    def set_description(self, description: str) -> None:
        if isinstance(self.parent, ColumnMetadata):
            LOGGER.warning("""Frontend does not currently support setting descriptions for type metadata
                           objects with a ColumnMetadata parent, since the top level type metadata does
                           not have its own row in the column table""")
        elif isinstance(self.parent, ArrayTypeMetadata):
            LOGGER.warning("""Frontend does not currently support setting descriptions for type metadata
                           objects with an ArrayTypeMetadata parent, since this level in the nesting
                           hierarchy is not named and therefore is represented by short row that is not
                           clickable""")
        else:
            self._description = DescriptionMetadata.create_description_metadata(
                source=None,
                text=description
            )

    def get_badges(self) -> Optional[List[Badge]]:
        return self._badges

    def set_badges(self, badges: Union[List[str], None] = None) -> None:
        if isinstance(self.parent, ColumnMetadata):
            LOGGER.warning("""Frontend does not currently support setting badges for type metadata
                           objects with a ColumnMetadata parent, since the top level type metadata does
                           not have its own row in the column table""")
        elif isinstance(self.parent, ArrayTypeMetadata):
            LOGGER.warning("""Frontend does not currently support setting badges for type metadata
                           objects with an ArrayTypeMetadata parent, since this level in the nesting
                           hierarchy is not named and therefore is represented by short row that is not
                           clickable""")
        else:
            formatted_badges = _format_as_list(badges)
            self._badges = [Badge(badge, 'type_metadata') for badge in formatted_badges]

    @abc.abstractmethod
    def __eq__(self, other: Any) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def is_terminal_type(self) -> bool:
        """
        This is used to determine whether any child nodes
        should be created for the associated TypeMetadata object.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def create_node_iterator(self) -> Iterator[GraphNode]:
        raise NotImplementedError

    @abc.abstractmethod
    def create_relation_iterator(self) -> Iterator[GraphRelationship]:
        raise NotImplementedError

    def create_next_node(self) -> Optional[GraphNode]:
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Optional[GraphRelationship]:
        try:
            return next(self._relation_iter)
        except StopIteration:
            return None

    def key(self) -> str:
        if isinstance(self.parent, ColumnMetadata):
            return f"{self.parent_key()}/type/{self.name}"
        return f"{self.parent_key()}/{self.name}"

    def description_key(self) -> Optional[str]:
        if self._description:
            description_id = self._description.get_description_id()
            return f"{self.key()}/{description_id}"
        return None

    def relation_type(self) -> str:
        if isinstance(self.parent, ColumnMetadata):
            return TypeMetadata.COL_TM_RELATION_TYPE
        return TypeMetadata.SUBTYPE_RELATION_TYPE

    def inverse_relation_type(self) -> str:
        if isinstance(self.parent, ColumnMetadata):
            return TypeMetadata.TM_COL_RELATION_TYPE
        return TypeMetadata.INVERSE_SUBTYPE_RELATION_TYPE

    def parent_key(self) -> str:
        if isinstance(self.parent, ColumnMetadata):
            column_key = self.parent.get_column_key()
            assert column_key is not None, f"Column key must be set for {self.parent.name}"
            return column_key
        return self.parent.key()

    def parent_label(self) -> str:
        if isinstance(self.parent, ColumnMetadata):
            return ColumnMetadata.COLUMN_NODE_LABEL
        return TypeMetadata.NODE_LABEL

    def __repr__(self) -> str:
        return f"TypeMetadata({self.type_str!r})"


class ArrayTypeMetadata(TypeMetadata):
    kind = 'array'

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(ArrayTypeMetadata, self).__init__(*args, **kwargs)
        self.array_inner_type: Optional[TypeMetadata] = None

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ArrayTypeMetadata):
            return (self.name == other.name and
                    self.type_str == other.type_str and
                    self.sort_order == other.sort_order and
                    self._description == other._description and
                    self._badges == other._badges and
                    self.array_inner_type == other.array_inner_type and
                    self.key() == other.key())
        return False

    def is_terminal_type(self) -> bool:
        return not self.array_inner_type

    def create_node_iterator(self) -> Iterator[GraphNode]:
        node_attributes: Dict[str, Union[str, None, int]] = {
            TypeMetadata.KIND: self.kind,
            TypeMetadata.NAME: self.name,
            TypeMetadata.DATA_TYPE: self.type_str
        }

        if isinstance(self.sort_order, int):
            node_attributes[TypeMetadata.SORT_ORDER] = self.sort_order

        yield GraphNode(
            key=self.key(),
            label=TypeMetadata.NODE_LABEL,
            attributes=node_attributes
        )

        if self._description:
            description_key = self.description_key()
            assert description_key is not None, f"Could not retrieve description key for {self.name}"
            yield self._description.get_node(description_key)

        if self._badges:
            badge_metadata = BadgeMetadata(start_label=TypeMetadata.NODE_LABEL,
                                           start_key=self.key(),
                                           badges=self._badges)
            badge_nodes = badge_metadata.get_badge_nodes()
            for node in badge_nodes:
                yield node

        if not self.is_terminal_type():
            assert self.array_inner_type is not None, f"Array inner type must be set for {self.name}"
            yield from self.array_inner_type.create_node_iterator()

    def create_relation_iterator(self) -> Iterator[GraphRelationship]:
        yield GraphRelationship(
            start_label=self.parent_label(),
            start_key=self.parent_key(),
            end_label=TypeMetadata.NODE_LABEL,
            end_key=self.key(),
            type=self.relation_type(),
            reverse_type=self.inverse_relation_type(),
            attributes={}
        )

        if self._description:
            description_key = self.description_key()
            assert description_key is not None, f"Could not retrieve description key for {self.name}"
            yield self._description.get_relation(
                TypeMetadata.NODE_LABEL,
                self.key(),
                description_key
            )

        if self._badges:
            badge_metadata = BadgeMetadata(start_label=TypeMetadata.NODE_LABEL,
                                           start_key=self.key(),
                                           badges=self._badges)
            badge_relations = badge_metadata.get_badge_relations()
            for relation in badge_relations:
                yield relation

        if not self.is_terminal_type():
            assert self.array_inner_type is not None, f"Array inner type must be set for {self.name}"
            yield from self.array_inner_type.create_relation_iterator()


class MapTypeMetadata(TypeMetadata):
    kind = 'map'

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(MapTypeMetadata, self).__init__(*args, **kwargs)
        self.map_key_type: Optional[TypeMetadata] = None
        self.map_value_type: Optional[TypeMetadata] = None

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, MapTypeMetadata):
            return (self.name == other.name and
                    self.map_key_type == other.map_key_type and
                    self.map_value_type == other.map_value_type and
                    self.type_str == other.type_str and
                    self.sort_order == other.sort_order and
                    self._description == other._description and
                    self._badges == other._badges and
                    self.key() == other.key())
        return False

    def is_terminal_type(self) -> bool:
        return not self.map_key_type or not self.map_value_type

    def create_node_iterator(self) -> Iterator[GraphNode]:
        node_attributes: Dict[str, Union[str, None, int]] = {
            TypeMetadata.KIND: self.kind,
            TypeMetadata.NAME: self.name,
            TypeMetadata.DATA_TYPE: self.type_str
        }

        if isinstance(self.sort_order, int):
            node_attributes[TypeMetadata.SORT_ORDER] = self.sort_order

        yield GraphNode(
            key=self.key(),
            label=TypeMetadata.NODE_LABEL,
            attributes=node_attributes
        )

        if self._description:
            description_key = self.description_key()
            assert description_key is not None, f"Could not retrieve description key for {self.name}"
            yield self._description.get_node(description_key)

        if self._badges:
            badge_metadata = BadgeMetadata(start_label=TypeMetadata.NODE_LABEL,
                                           start_key=self.key(),
                                           badges=self._badges)
            badge_nodes = badge_metadata.get_badge_nodes()
            for node in badge_nodes:
                yield node

        if not self.is_terminal_type():
            assert self.map_key_type is not None, f"Map key type must be set for {self.name}"
            assert self.map_value_type is not None, f"Map value type must be set for {self.name}"
            yield from self.map_key_type.create_node_iterator()
            yield from self.map_value_type.create_node_iterator()

    def create_relation_iterator(self) -> Iterator[GraphRelationship]:
        yield GraphRelationship(
            start_label=self.parent_label(),
            start_key=self.parent_key(),
            end_label=TypeMetadata.NODE_LABEL,
            end_key=self.key(),
            type=self.relation_type(),
            reverse_type=self.inverse_relation_type(),
            attributes={}
        )

        if self._description:
            description_key = self.description_key()
            assert description_key is not None, f"Could not retrieve description key for {self.name}"
            yield self._description.get_relation(
                TypeMetadata.NODE_LABEL,
                self.key(),
                description_key
            )

        if self._badges:
            badge_metadata = BadgeMetadata(start_label=TypeMetadata.NODE_LABEL,
                                           start_key=self.key(),
                                           badges=self._badges)
            badge_relations = badge_metadata.get_badge_relations()
            for relation in badge_relations:
                yield relation

        if not self.is_terminal_type():
            assert self.map_key_type is not None, f"Map key type must be set for {self.name}"
            assert self.map_value_type is not None, f"Map value type must be set for {self.name}"
            yield from self.map_key_type.create_relation_iterator()
            yield from self.map_value_type.create_relation_iterator()


class ScalarTypeMetadata(TypeMetadata):
    """
    ScalarTypeMetadata represents any non complex type that does not
    require special handling. It is also used as the default TypeMetadata
    class when a type string cannot be parsed.
    """
    kind = 'scalar'

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(ScalarTypeMetadata, self).__init__(*args, **kwargs)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ScalarTypeMetadata):
            return (self.name == other.name and
                    self.type_str == other.type_str and
                    self.sort_order == other.sort_order and
                    self._description == other._description and
                    self._badges == other._badges and
                    self.key() == other.key())
        return False

    def is_terminal_type(self) -> bool:
        return True

    def create_node_iterator(self) -> Iterator[GraphNode]:
        node_attributes: Dict[str, Union[str, None, int]] = {
            TypeMetadata.KIND: self.kind,
            TypeMetadata.NAME: self.name,
            TypeMetadata.DATA_TYPE: self.type_str
        }

        if isinstance(self.sort_order, int):
            node_attributes[TypeMetadata.SORT_ORDER] = self.sort_order

        yield GraphNode(
            key=self.key(),
            label=TypeMetadata.NODE_LABEL,
            attributes=node_attributes
        )

        if self._description:
            description_key = self.description_key()
            assert description_key is not None, f"Could not retrieve description key for {self.name}"
            yield self._description.get_node(description_key)

        if self._badges:
            badge_metadata = BadgeMetadata(start_label=TypeMetadata.NODE_LABEL,
                                           start_key=self.key(),
                                           badges=self._badges)
            badge_nodes = badge_metadata.get_badge_nodes()
            for node in badge_nodes:
                yield node

    def create_relation_iterator(self) -> Iterator[GraphRelationship]:
        yield GraphRelationship(
            start_label=self.parent_label(),
            start_key=self.parent_key(),
            end_label=TypeMetadata.NODE_LABEL,
            end_key=self.key(),
            type=self.relation_type(),
            reverse_type=self.inverse_relation_type(),
            attributes={}
        )

        if self._description:
            description_key = self.description_key()
            assert description_key is not None, f"Could not retrieve description key for {self.name}"
            yield self._description.get_relation(
                TypeMetadata.NODE_LABEL,
                self.key(),
                description_key
            )

        if self._badges:
            badge_metadata = BadgeMetadata(start_label=TypeMetadata.NODE_LABEL,
                                           start_key=self.key(),
                                           badges=self._badges)
            badge_relations = badge_metadata.get_badge_relations()
            for relation in badge_relations:
                yield relation


class StructTypeMetadata(TypeMetadata):
    kind = 'struct'

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(StructTypeMetadata, self).__init__(*args, **kwargs)
        self.struct_items: Optional[Dict[str, TypeMetadata]] = None

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, StructTypeMetadata):
            return (self.name == other.name and
                    self.struct_items == other.struct_items and
                    self.type_str == other.type_str and
                    self.sort_order == other.sort_order and
                    self._description == other._description and
                    self._badges == other._badges and
                    self.key() == other.key())
        return False

    def is_terminal_type(self) -> bool:
        return not self.struct_items

    def create_node_iterator(self) -> Iterator[GraphNode]:
        node_attributes: Dict[str, Union[str, None, int]] = {
            TypeMetadata.KIND: self.kind,
            TypeMetadata.NAME: self.name,
            TypeMetadata.DATA_TYPE: self.type_str
        }

        if isinstance(self.sort_order, int):
            node_attributes[TypeMetadata.SORT_ORDER] = self.sort_order

        yield GraphNode(
            key=self.key(),
            label=TypeMetadata.NODE_LABEL,
            attributes=node_attributes
        )

        if self._description:
            description_key = self.description_key()
            assert description_key is not None, f"Could not retrieve description key for {self.name}"
            yield self._description.get_node(description_key)

        if self._badges:
            badge_metadata = BadgeMetadata(start_label=TypeMetadata.NODE_LABEL,
                                           start_key=self.key(),
                                           badges=self._badges)
            badge_nodes = badge_metadata.get_badge_nodes()
            for node in badge_nodes:
                yield node

        if not self.is_terminal_type():
            assert self.struct_items, f"Struct items must be set for {self.name}"
            for name, data_type in self.struct_items.items():
                yield from data_type.create_node_iterator()

    def create_relation_iterator(self) -> Iterator[GraphRelationship]:
        yield GraphRelationship(
            start_label=self.parent_label(),
            start_key=self.parent_key(),
            end_label=TypeMetadata.NODE_LABEL,
            end_key=self.key(),
            type=self.relation_type(),
            reverse_type=self.inverse_relation_type(),
            attributes={}
        )

        if self._description:
            description_key = self.description_key()
            assert description_key is not None, f"Could not retrieve description key for {self.name}"
            yield self._description.get_relation(
                TypeMetadata.NODE_LABEL,
                self.key(),
                description_key
            )

        if self._badges:
            badge_metadata = BadgeMetadata(start_label=TypeMetadata.NODE_LABEL,
                                           start_key=self.key(),
                                           badges=self._badges)
            badge_relations = badge_metadata.get_badge_relations()
            for relation in badge_relations:
                yield relation

        if not self.is_terminal_type():
            assert self.struct_items, f"Struct items must be set for {self.name}"
            for name, data_type in self.struct_items.items():
                yield from data_type.create_relation_iterator()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Iterator, Union

from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor import sql_alchemy_extractor
from databuilder.extractor.base_extractor import Extractor
from databuilder.models.table_last_updated import TableLastUpdated

LOGGER = logging.getLogger(__name__)


class SnowflakeTableLastUpdatedExtractor(Extractor):
    """
    Extracts Snowflake table last update time from INFORMATION_SCHEMA metadata tables using SQLAlchemyExtractor.
    Requirements:
        snowflake-connector-python
        snowflake-sqlalchemy
    """
    # https://docs.snowflake.com/en/sql-reference/info-schema/views.html#columns
    # 'last_altered' column in 'TABLES` metadata view under 'INFORMATION_SCHEMA' contains last time when the table was
    # updated (both DML and DDL update). Below query fetches that column for each table.
    SQL_STATEMENT = """
        SELECT
            lower({cluster_source}) AS cluster,
            lower(t.table_schema) AS schema,
            lower(t.table_name) AS table_name,
            DATE_PART(EPOCH, t.last_altered) AS last_updated_time
        FROM
            {database}.INFORMATION_SCHEMA.TABLES t
        {where_clause_suffix};
        """

    # CONFIG KEYS
    WHERE_CLAUSE_SUFFIX_KEY = 'where_clause_suffix'
    CLUSTER_KEY = 'cluster_key'
    USE_CATALOG_AS_CLUSTER_NAME = 'use_catalog_as_cluster_name'
    # Database Key, used to identify the database type in the UI.
    DATABASE_KEY = 'database_key'
    # Snowflake Database Key, used to determine which Snowflake database to connect to.
    SNOWFLAKE_DATABASE_KEY = 'snowflake_database'

    # Default values
    DEFAULT_CLUSTER_NAME = 'master'

    DEFAULT_CONFIG = ConfigFactory.from_dict(
        {WHERE_CLAUSE_SUFFIX_KEY: ' WHERE t.last_altered IS NOT NULL ',
         CLUSTER_KEY: DEFAULT_CLUSTER_NAME,
         USE_CATALOG_AS_CLUSTER_NAME: True,
         DATABASE_KEY: 'snowflake',
         SNOWFLAKE_DATABASE_KEY: 'prod'}
    )

    def init(self, conf: ConfigTree) -> None:
        conf = conf.with_fallback(SnowflakeTableLastUpdatedExtractor.DEFAULT_CONFIG)
        self._cluster = conf.get_string(SnowflakeTableLastUpdatedExtractor.CLUSTER_KEY)

        if conf.get_bool(SnowflakeTableLastUpdatedExtractor.USE_CATALOG_AS_CLUSTER_NAME):
            cluster_source = "t.table_catalog"
        else:
            cluster_source = f"'{self._cluster}'"

        self._database = conf.get_string(SnowflakeTableLastUpdatedExtractor.DATABASE_KEY)
        self._snowflake_database = conf.get_string(SnowflakeTableLastUpdatedExtractor.SNOWFLAKE_DATABASE_KEY)

        self.sql_stmt = SnowflakeTableLastUpdatedExtractor.SQL_STATEMENT.format(
            where_clause_suffix=conf.get_string(SnowflakeTableLastUpdatedExtractor.WHERE_CLAUSE_SUFFIX_KEY),
            cluster_source=cluster_source,
            database=self._snowflake_database
        )

        LOGGER.info('SQL for snowflake table last updated timestamp: %s', self.sql_stmt)

        # use an sql_alchemy_extractor to execute sql
        self._alchemy_extractor = sql_alchemy_extractor.from_surrounding_config(conf, self.sql_stmt)
        self._extract_iter: Union[None, Iterator] = None

    def close(self) -> None:
        if getattr(self, '_alchemy_extractor', None) is not None:
            self._alchemy_extractor.close()

    def extract(self) -> Union[TableLastUpdated, None]:
        if not self._extract_iter:
            self._extract_iter = self._get_extract_iter()
        try:
            return next(self._extract_iter)
        except StopIteration:
            return None

    def get_scope(self) -> str:
        return 'extractor.snowflake_table_last_updated'

    def _get_extract_iter(self) -> Iterator[TableLastUpdated]:
        """
        Provides iterator of result row from SQLAlchemy extractor
        """
        tbl_last_updated_row = self._alchemy_extractor.extract()
        while tbl_last_updated_row:
            yield TableLastUpdated(table_name=tbl_last_updated_row['table_name'],
                                   last_updated_time_epoch=tbl_last_updated_row['last_updated_time'],
                                   schema=tbl_last_updated_row['schema'],
                                   db=self._database,
                                   cluster=tbl_last_updated_row['cluster'])
            tbl_last_updated_row = self._alchemy_extractor.extract()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import atexit
from typing import Callable, List


class Closer(object):
    """
    A Closer class that is responsible for collecting closeable callable,
    and close it in group. Registered closeable callable will be guaranteed
    to be called where only last occurred failure will be propagated back.

    Order of closing registered closeable callable will be LIFO
    as closeable instance can have dependency each other.
    """

    def __init__(self) -> None:
        self._stack: List = []
        atexit.register(self.close)

    def register(self, close_callable: Callable) -> None:
        """
        Register closeable callable.
        :param close_callable:
        :return: None
        """
        if not callable(close_callable):
            raise RuntimeError(f'Only callable can be registered: {close_callable}')

        self._stack.append(close_callable)

    def close(self) -> None:
        """
        Execute all closeable callable in LIFO order.
        All registered callable will be guaranteed to be executed. If there
        are multiple failures, only the last one will be propagated.
        :return:
        """
        if not self._stack:
            return

        last_exception = None
        while len(self._stack):
            try:
                close_callable = self._stack.pop()
                close_callable()
            except Exception as e:
                last_exception = e

        if last_exception:
            raise last_exception
